from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer

# hex: 0x000B00
_REQUEST_MESSAGE_TYPE = 2816
# hex: 0x000B01
_RESPONSE_MESSAGE_TYPE = 2817

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request():
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    return OutboundMessage(buf, True)



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.sql import _SqlQueryId

_MEMBER_ID_HIGH_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_MEMBER_ID_HIGH_DECODE_OFFSET = 0
_MEMBER_ID_LOW_ENCODE_OFFSET = _MEMBER_ID_HIGH_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_MEMBER_ID_LOW_DECODE_OFFSET = _MEMBER_ID_HIGH_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_LOCAL_ID_HIGH_ENCODE_OFFSET = _MEMBER_ID_LOW_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_LOCAL_ID_HIGH_DECODE_OFFSET = _MEMBER_ID_LOW_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_LOCAL_ID_LOW_ENCODE_OFFSET = _LOCAL_ID_HIGH_ENCODE_OFFSET + LONG_SIZE_IN_BYTES
_LOCAL_ID_LOW_DECODE_OFFSET = _LOCAL_ID_HIGH_DECODE_OFFSET + LONG_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _LOCAL_ID_LOW_ENCODE_OFFSET + LONG_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class SqlQueryIdCodec(object):
    @staticmethod
    def encode(buf, sql_query_id, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _MEMBER_ID_HIGH_ENCODE_OFFSET, sql_query_id.member_id_high)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _MEMBER_ID_LOW_ENCODE_OFFSET, sql_query_id.member_id_low)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _LOCAL_ID_HIGH_ENCODE_OFFSET, sql_query_id.local_id_high)
        FixSizedTypesCodec.encode_long(initial_frame_buf, _LOCAL_ID_LOW_ENCODE_OFFSET, sql_query_id.local_id_low)
        buf.extend(initial_frame_buf)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        member_id_high = FixSizedTypesCodec.decode_long(initial_frame.buf, _MEMBER_ID_HIGH_DECODE_OFFSET)
        member_id_low = FixSizedTypesCodec.decode_long(initial_frame.buf, _MEMBER_ID_LOW_DECODE_OFFSET)
        local_id_high = FixSizedTypesCodec.decode_long(initial_frame.buf, _LOCAL_ID_HIGH_DECODE_OFFSET)
        local_id_low = FixSizedTypesCodec.decode_long(initial_frame.buf, _LOCAL_ID_LOW_DECODE_OFFSET)
        CodecUtil.fast_forward_to_end_frame(msg)
        return _SqlQueryId(member_id_high, member_id_low, local_id_high, local_id_low)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import ByteArrayCodec

# hex: 0x000C00
_REQUEST_MESSAGE_TYPE = 3072
# hex: 0x000C01
_RESPONSE_MESSAGE_TYPE = 3073

_REQUEST_TIMESTAMP_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_TIMESTAMP_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(timestamp, client_attributes, metrics_blob):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_TIMESTAMP_OFFSET, timestamp)
    StringCodec.encode(buf, client_attributes)
    ByteArrayCodec.encode(buf, metrics_blob, True)
    return OutboundMessage(buf, False)



import logging
import time
import functools

from hazelcast.errors import (
    create_error_from_message,
    HazelcastInstanceNotActiveError,
    is_retryable_error,
    TargetDisconnectedError,
    HazelcastClientNotActiveError,
    TargetNotMemberError,
    EXCEPTION_MESSAGE_TYPE,
    IndeterminateOperationStateError,
    OperationTimeoutError,
)
from hazelcast.future import Future
from hazelcast.protocol.codec import client_local_backup_listener_codec
from hazelcast.util import AtomicInteger
from hazelcast import six

_logger = logging.getLogger(__name__)


def _no_op_response_handler(_):
    pass


class Invocation(object):
    __slots__ = (
        "request",
        "timeout",
        "partition_id",
        "uuid",
        "connection",
        "event_handler",
        "future",
        "sent_connection",
        "urgent",
        "response_handler",
        "backup_acks_received",
        "backup_acks_expected",
        "pending_response",
        "pending_response_received_time",
    )

    def __init__(
        self,
        request,
        partition_id=-1,
        uuid=None,
        connection=None,
        event_handler=None,
        urgent=False,
        timeout=None,
        response_handler=_no_op_response_handler,
    ):
        self.request = request
        self.partition_id = partition_id
        self.uuid = uuid
        self.connection = connection
        self.event_handler = event_handler
        self.urgent = urgent
        self.timeout = timeout
        self.future = Future()
        self.sent_connection = None
        self.response_handler = response_handler
        self.backup_acks_received = 0
        self.backup_acks_expected = -1
        self.pending_response = None
        self.pending_response_received_time = -1

    def set_response(self, response):
        try:
            result = self.response_handler(response)
            self.future.set_result(result)
        except Exception as e:
            self.future.set_exception(e)

    def set_exception(self, exception, traceback=None):
        self.future.set_exception(exception, traceback)


class InvocationService(object):
    _CLEAN_RESOURCES_PERIOD = 0.1

    def __init__(self, client, config, reactor):
        smart_routing = config.smart_routing
        if smart_routing:
            self._do_invoke = self._invoke_smart
        else:
            self._do_invoke = self._invoke_non_smart

        self._client = client
        self._reactor = reactor
        self._partition_service = None
        self._connection_manager = None
        self._listener_service = None
        self._check_invocation_allowed_fn = None
        self._pending = {}
        self._next_correlation_id = AtomicInteger(1)
        self._is_redo_operation = config.redo_operation
        self._invocation_timeout = config.invocation_timeout
        self._invocation_retry_pause = config.invocation_retry_pause
        self._backup_ack_to_client_enabled = smart_routing and config.backup_ack_to_client_enabled
        self._fail_on_indeterminate_state = config.fail_on_indeterminate_operation_state
        self._backup_timeout = config.operation_backup_timeout
        self._clean_resources_timer = None
        self._shutdown = False

    def init(self, partition_service, connection_manager, listener_service):
        self._partition_service = partition_service
        self._connection_manager = connection_manager
        self._listener_service = listener_service
        self._check_invocation_allowed_fn = connection_manager.check_invocation_allowed

    def start(self):
        self._start_clean_resources_timer()

    def add_backup_listener(self):
        if self._backup_ack_to_client_enabled:
            self._register_backup_listener()

    def handle_client_message(self, message):
        correlation_id = message.get_correlation_id()

        start_frame = message.start_frame
        if start_frame.has_event_flag() or start_frame.has_backup_event_flag():
            self._listener_service.handle_client_message(message, correlation_id)
            return

        invocation = self._pending.get(correlation_id, None)
        if not invocation:
            _logger.warning("Got message with unknown correlation id: %s", message)
            return

        if message.get_message_type() == EXCEPTION_MESSAGE_TYPE:
            error = create_error_from_message(message)
            return self._notify_error(invocation, error)

        self._notify(invocation, message)

    def invoke(self, invocation):
        if not invocation.timeout:
            invocation.timeout = self._invocation_timeout + time.time()

        correlation_id = self._next_correlation_id.get_and_increment()
        request = invocation.request
        request.set_correlation_id(correlation_id)
        request.set_partition_id(invocation.partition_id)
        self._do_invoke(invocation)

    def shutdown(self):
        if self._shutdown:
            return

        self._shutdown = True
        if self._clean_resources_timer:
            self._clean_resources_timer.cancel()
        for invocation in list(six.itervalues(self._pending)):
            self._notify_error(invocation, HazelcastClientNotActiveError())

    def _invoke_on_partition_owner(self, invocation, partition_id):
        owner_uuid = self._partition_service.get_partition_owner(partition_id)
        if not owner_uuid:
            _logger.debug("Partition owner is not assigned yet")
            return False
        return self._invoke_on_target(invocation, owner_uuid)

    def _invoke_on_target(self, invocation, owner_uuid):
        connection = self._connection_manager.get_connection(owner_uuid)
        if not connection:
            _logger.debug("Client is not connected to target: %s", owner_uuid)
            return False
        return self._send(invocation, connection)

    def _invoke_on_random_connection(self, invocation):
        connection = self._connection_manager.get_random_connection()
        if not connection:
            _logger.debug("No connection found to invoke")
            return False
        return self._send(invocation, connection)

    def _invoke_smart(self, invocation):
        try:
            if not invocation.urgent:
                self._check_invocation_allowed_fn()

            connection = invocation.connection
            if connection:
                invoked = self._send(invocation, connection)
                if not invoked:
                    self._notify_error(
                        invocation, IOError("Could not invoke on connection %s" % connection)
                    )
                return

            if invocation.partition_id != -1:
                invoked = self._invoke_on_partition_owner(invocation, invocation.partition_id)
            elif invocation.uuid:
                invoked = self._invoke_on_target(invocation, invocation.uuid)
            else:
                invoked = self._invoke_on_random_connection(invocation)

            if not invoked:
                invoked = self._invoke_on_random_connection(invocation)

            if not invoked:
                self._notify_error(invocation, IOError("No connection found to invoke"))
        except Exception as e:
            self._notify_error(invocation, e)

    def _invoke_non_smart(self, invocation):
        try:
            if not invocation.urgent:
                self._check_invocation_allowed_fn()

            connection = invocation.connection
            if connection:
                invoked = self._send(invocation, connection)
                if not invoked:
                    self._notify_error(
                        invocation, IOError("Could not invoke on connection %s" % connection)
                    )
                return

            if not self._invoke_on_random_connection(invocation):
                self._notify_error(invocation, IOError("No connection found to invoke"))
        except Exception as e:
            self._notify_error(invocation, e)

    def _send(self, invocation, connection):
        if self._shutdown:
            raise HazelcastClientNotActiveError()

        if self._backup_ack_to_client_enabled:
            invocation.request.set_backup_aware_flag()

        message = invocation.request
        correlation_id = message.get_correlation_id()
        self._pending[correlation_id] = invocation

        if invocation.event_handler:
            self._listener_service.add_event_handler(correlation_id, invocation.event_handler)

        if not connection.send_message(message):
            if invocation.event_handler:
                self._listener_service.remove_event_handler(correlation_id)
            return False

        invocation.sent_connection = connection
        return True

    def _notify_error(self, invocation, error):
        _logger.debug("Got exception for request %s, error: %s", invocation.request, error)

        if not self._client.lifecycle_service.is_running():
            self._complete_with_error(invocation, HazelcastClientNotActiveError())
            return

        if not self._should_retry(invocation, error):
            self._complete_with_error(invocation, error)
            return

        if invocation.timeout < time.time():
            _logger.debug("Error will not be retried because invocation timed out: %s", error)
            error = OperationTimeoutError(
                "Request timed out because an error occurred "
                "after invocation timeout: %s" % error
            )
            self._complete_with_error(invocation, error)
            return

        invocation.sent_connection = None
        invoke_func = functools.partial(self._retry_if_not_done, invocation)
        self._reactor.add_timer(self._invocation_retry_pause, invoke_func)

    def _retry_if_not_done(self, invocation):
        if not invocation.future.done():
            self._do_invoke(invocation)

    def _should_retry(self, invocation, error):
        if invocation.connection and isinstance(error, (IOError, TargetDisconnectedError)):
            return False

        if invocation.uuid and isinstance(error, TargetNotMemberError):
            return False

        if isinstance(error, (IOError, HazelcastInstanceNotActiveError)) or is_retryable_error(
            error
        ):
            return True

        if isinstance(error, TargetDisconnectedError):
            return invocation.request.retryable or self._is_redo_operation

        return False

    def _complete_with_error(self, invocation, error):
        invocation.set_exception(error, None)
        correlation_id = invocation.request.get_correlation_id()
        self._pending.pop(correlation_id, None)

    def _register_backup_listener(self):
        codec = client_local_backup_listener_codec
        request = codec.encode_request()
        self._listener_service.register_listener(
            request,
            codec.decode_response,
            lambda reg_id: None,
            lambda m: codec.handle(m, self._backup_event_handler),
        ).result()

    def _backup_event_handler(self, correlation_id):
        invocation = self._pending.get(correlation_id, None)
        if not invocation:
            _logger.debug("Invocation not found for backup event, invocation id %s", correlation_id)
            return
        self._notify_backup_complete(invocation)

    def _notify(self, invocation, client_message):
        expected_backups = client_message.get_number_of_backup_acks()
        if expected_backups > invocation.backup_acks_received:
            invocation.pending_response_received_time = time.time()
            invocation.backup_acks_expected = expected_backups
            invocation.pending_response = client_message
            return

        self._complete(invocation, client_message)

    def _notify_backup_complete(self, invocation):
        invocation.backup_acks_received += 1
        if not invocation.pending_response:
            return

        if invocation.backup_acks_expected != invocation.backup_acks_received:
            return

        self._complete(invocation, invocation.pending_response)

    def _complete(self, invocation, client_message):
        invocation.set_response(client_message)
        correlation_id = invocation.request.get_correlation_id()
        self._pending.pop(correlation_id, None)

    def _start_clean_resources_timer(self):
        def run():
            if self._shutdown:
                return

            now = time.time()
            for invocation in list(self._pending.values()):
                connection = invocation.sent_connection
                if not connection:
                    continue

                if not connection.live:
                    error = TargetDisconnectedError(connection.close_reason)
                    self._notify_error(invocation, error)
                    continue

                if self._backup_ack_to_client_enabled:
                    self._detect_and_handle_backup_timeout(invocation, now)

            self._clean_resources_timer = self._reactor.add_timer(self._CLEAN_RESOURCES_PERIOD, run)

        self._clean_resources_timer = self._reactor.add_timer(self._CLEAN_RESOURCES_PERIOD, run)

    def _detect_and_handle_backup_timeout(self, invocation, now):
        if not invocation.pending_response:
            return

        if invocation.backup_acks_expected == invocation.backup_acks_received:
            return

        expiration_time = invocation.pending_response_received_time + self._backup_timeout
        timeout_reached = 0 < expiration_time < now
        if not timeout_reached:
            return

        if self._fail_on_indeterminate_state:
            error = IndeterminateOperationStateError(
                "Invocation failed because the backup acks are missed"
            )
            self._complete_with_error(invocation, error)
            return

        self._complete(invocation, invocation.pending_response)



from hazelcast.protocol.codec import (
    transactional_queue_offer_codec,
    transactional_queue_peek_codec,
    transactional_queue_poll_codec,
    transactional_queue_size_codec,
    transactional_queue_take_codec,
)
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none, to_millis, thread_id


class TransactionalQueue(TransactionalProxy):
    """Transactional implementation of :class:`~hazelcast.proxy.queue.Queue`."""

    def offer(self, item, timeout=0):
        """Transactional implementation of :func:`Queue.offer(item, timeout) <hazelcast.proxy.queue.Queue.offer>`

        Args:
            item: The item to be added.
            timeout (int): Maximum time in seconds to wait for addition.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the element was added to this queue, ``False`` otherwise.
        """
        check_not_none(item, "item can't be none")

        item_data = self._to_data(item)
        request = transactional_queue_offer_codec.encode_request(
            self.name, self.transaction.id, thread_id(), item_data, to_millis(timeout)
        )
        return self._invoke(request, transactional_queue_offer_codec.decode_response)

    def take(self):
        """Transactional implementation of :func:`Queue.take() <hazelcast.proxy.queue.Queue.take>`

        Returns:
            hazelcast.future.Future[any]: The head of this queue.
        """

        def handler(message):
            return self._to_object(transactional_queue_take_codec.decode_response(message))

        request = transactional_queue_take_codec.encode_request(
            self.name, self.transaction.id, thread_id()
        )
        return self._invoke(request, handler)

    def poll(self, timeout=0):
        """Transactional implementation of :func:`Queue.poll(timeout) <hazelcast.proxy.queue.Queue.poll>`

        Args:
            timeout (int): Maximum time in seconds to wait for addition.

        Returns:
            hazelcast.future.Future[any]: The head of this queue, or ``None`` if this queue is empty
            or specified timeout elapses before an item is added to the queue.
        """

        def handler(message):
            return self._to_object(transactional_queue_poll_codec.decode_response(message))

        request = transactional_queue_poll_codec.encode_request(
            self.name, self.transaction.id, thread_id(), to_millis(timeout)
        )
        return self._invoke(request, handler)

    def peek(self, timeout=0):
        """Transactional implementation of :func:`Queue.peek(timeout) <hazelcast.proxy.queue.Queue.peek>`

        Args:
            timeout (int): Maximum time in seconds to wait for addition.

        Returns:
            hazelcast.future.Future[any]: The head of this queue, or ``None`` if this queue is empty
            or specified timeout elapses before an item is added to the queue.
        """

        def handler(message):
            return self._to_object(transactional_queue_peek_codec.decode_response(message))

        request = transactional_queue_peek_codec.encode_request(
            self.name, self.transaction.id, thread_id(), to_millis(timeout)
        )
        return self._invoke(request, handler)

    def size(self):
        """Transactional implementation of :func:`Queue.size() <hazelcast.proxy.queue.Queue.size>`

        Returns:
            hazelcast.future.Future[int]: Size of the queue.
        """
        request = transactional_queue_size_codec.encode_request(
            self.name, self.transaction.id, thread_id()
        )
        return self._invoke(request, transactional_queue_size_codec.decode_response)



import binascii
import random
import threading
import time
import uuid

from hazelcast.serialization import UUID_MSB_SHIFT, UUID_LSB_MASK, UUID_MSB_MASK

try:
    from collections.abc import Sequence, Iterable
except ImportError:
    # Compatibility for Python2
    from collections import Sequence, Iterable

from hazelcast import six

DEFAULT_ADDRESS = "127.0.0.1"
DEFAULT_PORT = 5701

MILLISECONDS_IN_SECONDS = 1000
NANOSECONDS_IN_SECONDS = 1e9


def check_not_none(val, message):
    if val is None:
        raise AssertionError(message)


def check_true(val, message):
    if not val:
        raise AssertionError(message)


def check_not_negative(val, message):
    if val < 0:
        raise AssertionError(message)


def check_not_empty(collection, message):
    if not collection:
        raise AssertionError(message)


def check_is_number(val, message="Number value expected"):
    if not isinstance(val, number_types):
        raise AssertionError(message)


def check_is_int(val, message="Int value expected"):
    if not isinstance(val, six.integer_types):
        raise AssertionError(message)


def current_time():
    return time.time()


def current_time_in_millis():
    return to_millis(current_time())


def thread_id():
    return threading.currentThread().ident


def to_millis(seconds):
    if seconds is None:
        return -1
    return int(seconds * MILLISECONDS_IN_SECONDS)


def to_nanos(seconds):
    return int(seconds * NANOSECONDS_IN_SECONDS)


def validate_type(_type):
    if not isinstance(_type, type):
        raise ValueError("Serializer should be an instance of %s" % _type.__name__)


def validate_serializer(serializer, _type):
    if not issubclass(serializer, _type):
        raise ValueError("Serializer should be an instance of %s" % _type.__name__)


class AtomicInteger(object):
    """An Integer which can work atomically."""

    def __init__(self, initial=0):
        self._mux = threading.RLock()
        self._counter = initial

    def get_and_increment(self):
        """Returns the current value and increment it.

        Returns:
            int: Current value of AtomicInteger.
        """
        with self._mux:
            res = self._counter
            self._counter += 1
            return res

    def get(self):
        """Returns the current value.

        Returns:
            int: The current value.
        """
        with self._mux:
            return self._counter

    def add(self, count):
        """Adds the given value to the current value.
        Args:
            count (int): The value to add.
        """
        with self._mux:
            self._counter += count


class ImmutableLazyDataList(Sequence):
    def __init__(self, list_data, to_object):
        super(ImmutableLazyDataList, self).__init__()
        self._list_data = list_data
        self._list_obj = [None] * len(self._list_data)
        self.to_object = to_object

    def __contains__(self, value):
        return super(ImmutableLazyDataList, self).__contains__(value)

    def __len__(self):
        return self._list_data.__len__()

    def __getitem__(self, index):
        val = self._list_obj[index]
        if not val:
            data = self._list_data[index]
            if isinstance(data, tuple):
                (key, value) = data
                self._list_obj[index] = (self.to_object(key), self.to_object(value))
            else:
                self._list_obj[index] = self.to_object(data)
        return self._list_obj[index]

    def __eq__(self, other):
        if not isinstance(other, Iterable):
            return False
        self._populate()
        return self._list_obj == other

    def _populate(self):
        for index, data in enumerate(self._list_data):
            if not self._list_obj[index]:
                self.__getitem__(index)

    def __repr__(self):
        self._populate()
        return str(self._list_obj)


# Serialization Utilities


def get_portable_version(portable, default_version):
    try:
        version = portable.get_class_version()
    except AttributeError:
        version = default_version
    return version


# Version utilities
UNKNOWN_VERSION = -1
MAJOR_VERSION_MULTIPLIER = 10000
MINOR_VERSION_MULTIPLIER = 100


def calculate_version(version_str):
    if not version_str:
        return UNKNOWN_VERSION

    main_parts = version_str.split("-")
    tokens = main_parts[0].split(".")

    if len(tokens) < 2:
        return UNKNOWN_VERSION

    try:
        major_coeff = int(tokens[0])
        minor_coeff = int(tokens[1])

        calculated_version = (
            major_coeff * MAJOR_VERSION_MULTIPLIER + minor_coeff * MINOR_VERSION_MULTIPLIER
        )

        if len(tokens) > 2:
            patch_coeff = int(tokens[2])
            calculated_version += patch_coeff

        return calculated_version
    except ValueError:
        return UNKNOWN_VERSION


def to_list(*args, **kwargs):
    return list(*args, **kwargs)


def to_signed(unsigned, bit_len):
    mask = (1 << bit_len) - 1
    if unsigned & (1 << (bit_len - 1)):
        return unsigned | ~mask
    return unsigned & mask


def get_attr_name(cls, value):
    for attr_name, attr_value in six.iteritems(vars(cls)):
        if attr_value == value:
            return attr_name
    return None


def _get_enum_value(cls, key):
    if isinstance(key, six.string_types):
        return getattr(cls, key, None)
    return None


def try_to_get_enum_value(value, enum_class):
    if get_attr_name(enum_class, value):
        # If the value given by the user corresponds
        # to value of the one of the enum members,
        # it is okay to set it directly
        return value
    else:
        # We couldn't find a enum member name for the
        # given value. Try to match the given config
        # option with enum member names and get value
        # associated with it.
        enum_value = _get_enum_value(enum_class, value)
        if enum_value is not None:
            return enum_value
        else:
            raise TypeError(
                "%s must be equal to one of the values or "
                "names of the members of the %s" % (value, enum_class.__name__)
            )


number_types = (six.integer_types, float)
none_type = type(None)


class LoadBalancer(object):
    """Load balancer allows you to send operations to one of a number of endpoints (Members).
    It is up to the implementation to use different load balancing policies.

    If the client is configured with smart routing,
    only the operations that are not key based will be routed to the endpoint
    """

    def init(self, cluster_service):
        """Initializes the load balancer.

        Args:
            cluster_service (hazelcast.cluster.ClusterService): The cluster service to select members from
        """
        raise NotImplementedError("init")

    def next(self):
        """Returns the next member to route to.

        Returns:
            hazelcast.core.MemberInfo: the next member or ``None`` if no member is available.
        """
        raise NotImplementedError("next")

    def next_data_member(self):
        """Returns the next data member to route to.

        Returns:
            hazelcast.core.MemberInfo: The next data member or
            ``None`` if no data member is available.
        """
        return None

    def can_get_next_data_member(self):
        """Returns whether this instance supports getting data members
         through a call to :func:`next_data_member`.

        Returns:
            bool: ``True`` if this instance supports getting data members.
        """
        return False


class _Members(object):
    __slots__ = ("members", "data_members")

    def __init__(self, members, data_members):
        self.members = members
        self.data_members = data_members


class _AbstractLoadBalancer(LoadBalancer):
    def __init__(self):
        self._cluster_service = None
        self._members = _Members([], [])

    def init(self, cluster_service):
        self._cluster_service = cluster_service
        cluster_service.add_listener(self._listener, self._listener, True)

    def next(self):
        members = self._members.members
        return self._next(members)

    def next_data_member(self):
        members = self._members.data_members
        return self._next(members)

    def can_get_next_data_member(self):
        return True

    def _listener(self, _):
        members = self._cluster_service.get_members()
        data_members = [member for member in members if not member.lite_member]

        self._members = _Members(members, data_members)

    def _next(self, members):
        raise NotImplementedError("_next")


class RoundRobinLB(_AbstractLoadBalancer):
    """A load balancer implementation that relies on using round robin
    to a next member to send a request to.

    Round robin is done based on best effort basis, the order of members for concurrent calls to
    the next() is not guaranteed.
    """

    def __init__(self):
        super(RoundRobinLB, self).__init__()
        self._idx = 0

    def _next(self, members):
        if not members:
            return None

        n = len(members)
        idx = self._idx % n
        self._idx += 1
        return members[idx]


class RandomLB(_AbstractLoadBalancer):
    """A load balancer that selects a random member to route to."""

    def _next(self, members):
        if not members:
            return None
        idx = random.randrange(0, len(members))
        return members[idx]


class IterationType(object):
    """To differentiate users selection on result collection on map-wide
    operations like ``entry_set``, ``key_set``, ``values`` etc.
    """

    KEY = 0
    """Iterate over keys"""

    VALUE = 1
    """Iterate over values"""

    ENTRY = 2
    """Iterate over entries"""


class UUIDUtil(object):
    @staticmethod
    def to_bits(value):
        i = value.int
        most_significant_bits = to_signed(i >> UUID_MSB_SHIFT, 64)
        least_significant_bits = to_signed(i & UUID_LSB_MASK, 64)
        return most_significant_bits, least_significant_bits

    @staticmethod
    def from_bits(most_significant_bits, least_significant_bits):
        return uuid.UUID(
            int=(
                ((most_significant_bits << UUID_MSB_SHIFT) & UUID_MSB_MASK)
                | (least_significant_bits & UUID_LSB_MASK)
            )
        )


if hasattr(int, "from_bytes"):

    def int_from_bytes(buffer):
        return int.from_bytes(buffer, "big", signed=True)


else:
    # Compatibility with Python 2
    def int_from_bytes(buffer):
        buffer = bytearray(buffer)
        if buffer[0] & 0x80:
            neg = bytearray()
            for c in buffer:
                neg.append(~c)
            return -1 * int(binascii.hexlify(neg), 16) - 1
        return int(binascii.hexlify(buffer), 16)


try:
    from datetime import timezone
except ImportError:
    from datetime import tzinfo, timedelta

    # There is no tzinfo implementation(timezone) in the
    # Python 2. Here we provide the bare minimum
    # to the user.
    class FixedOffsetTimezone(tzinfo):
        __slots__ = ("_offset",)

        def __init__(self, offset):
            self._offset = offset

        def utcoffset(self, dt):
            return self._offset

        def tzname(self, dt):
            return None

        def dst(self, dt):
            return timedelta(0)

    timezone = FixedOffsetTimezone


def try_to_get_error_message(error):
    # If the error has a message attribute,
    # return it. If not, almost all of the
    # built-in errors (and Hazelcast Errors)
    # set the exception message as the first
    # parameter of args. If it is not there,
    # then return None.
    if hasattr(error, "message"):
        return error.message
    elif len(error.args) > 0:
        return error.args[0]
    return None



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec
from hazelcast.protocol.builtin import StringCodec

# hex: 0x070400
_REQUEST_MESSAGE_TYPE = 459776
# hex: 0x070401
_RESPONSE_MESSAGE_TYPE = 459777

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE
_RESPONSE_FENCE_OFFSET = RESPONSE_HEADER_SIZE
_RESPONSE_LOCK_COUNT_OFFSET = _RESPONSE_FENCE_OFFSET + LONG_SIZE_IN_BYTES
_RESPONSE_SESSION_ID_OFFSET = _RESPONSE_LOCK_COUNT_OFFSET + INT_SIZE_IN_BYTES
_RESPONSE_THREAD_ID_OFFSET = _RESPONSE_SESSION_ID_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(group_id, name):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    RaftGroupIdCodec.encode(buf, group_id)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["fence"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_FENCE_OFFSET)
    response["lock_count"] = FixSizedTypesCodec.decode_int(initial_frame.buf, _RESPONSE_LOCK_COUNT_OFFSET)
    response["session_id"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_SESSION_ID_OFFSET)
    response["thread_id"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_THREAD_ID_OFFSET)
    return response



import sys
from threading import RLock

from hazelcast.config import IntType
from hazelcast.serialization.api import *
from hazelcast.serialization.data import *
from hazelcast.errors import HazelcastInstanceNotActiveError, HazelcastSerializationError
from hazelcast.serialization.input import _ObjectDataInput
from hazelcast.serialization.output import _ObjectDataOutput
from hazelcast.serialization.serializer import *
from hazelcast import six


_int_type_to_type_id = {
    IntType.BYTE: CONSTANT_TYPE_BYTE,
    IntType.SHORT: CONSTANT_TYPE_SHORT,
    IntType.INT: CONSTANT_TYPE_INTEGER,
    IntType.LONG: CONSTANT_TYPE_LONG,
    IntType.BIG_INT: JAVA_DEFAULT_TYPE_BIG_INTEGER,
}


def empty_partitioning_strategy(_):
    return None


def handle_exception(e, traceback):
    if isinstance(e, MemoryError):
        # TODO
        six.print_("OUT OF MEMORY")
        six.reraise(MemoryError, e, traceback)
    elif isinstance(e, HazelcastSerializationError):
        six.reraise(HazelcastSerializationError, e, traceback)
    else:
        six.reraise(HazelcastSerializationError, HazelcastSerializationError(e.args[0]), traceback)


def is_null_data(data):
    return data.data_size() == 0 and data.get_type() == CONSTANT_TYPE_NULL


def index_for_default_type(type_id):
    return -type_id


class BaseSerializationService(object):
    def __init__(
        self, version, global_partition_strategy, output_buffer_size, is_big_endian, int_type
    ):
        self._registry = SerializerRegistry(int_type)
        self._version = version
        self._global_partition_strategy = global_partition_strategy
        self._output_buffer_size = output_buffer_size
        self._is_big_endian = is_big_endian
        self._active = True

    def to_data(self, obj, partitioning_strategy=None):
        """Serialize the input object into byte array representation

        Args:
            obj: Input object
            partitioning_strategy (function): Function in the form of ``lambda key: partitioning_key``.

        Returns:
            hazelcast.serialization.data.Data: Data object
        """
        if obj is None:
            return None

        if isinstance(obj, Data):
            return obj

        out = self._create_data_output()
        try:
            serializer = self._registry.serializer_for(obj)
            partitioning_hash = self._calculate_partitioning_hash(obj, partitioning_strategy)

            out.write_int_big_endian(partitioning_hash)
            out.write_int_big_endian(serializer.get_type_id())
            serializer.write(out, obj)
            return Data(out.to_byte_array())
        except:
            handle_exception(sys.exc_info()[1], sys.exc_info()[2])
        finally:
            pass
            # return out to pool

    def to_object(self, data):
        """Deserialize input data

        Args:
            data (hazelcast.serialization.data.Data): Serialized input Data object

        Returns:
            any: Deserialized object
        """
        if not isinstance(data, Data):
            return data

        if is_null_data(data):
            return None

        inp = self._create_data_input(data)
        try:
            type_id = data.get_type()
            serializer = self._registry.serializer_by_type_id(type_id)
            if serializer is None:
                if self._active:
                    raise HazelcastSerializationError("Missing Serializer for type-id:%s" % type_id)
                else:
                    raise HazelcastInstanceNotActiveError()
            return serializer.read(inp)
        except:
            handle_exception(sys.exc_info()[1], sys.exc_info()[2])
        finally:
            pass
            # return out to pool

    def write_object(self, out, obj):
        if isinstance(obj, Data):
            raise HazelcastSerializationError(
                "Cannot write a Data instance! Use write_data(out, data) instead."
            )
        try:
            serializer = self._registry.serializer_for(obj)
            out.write_int(serializer.get_type_id())
            serializer.write(out, obj)
        except:
            handle_exception(sys.exc_info()[1], sys.exc_info()[2])

    def read_object(self, inp):
        try:
            type_id = inp.read_int()
            serializer = self._registry.serializer_by_type_id(type_id)
            if serializer is None:
                if self._active:
                    raise HazelcastSerializationError(
                        "Missing Serializer for type-id: %s" % type_id
                    )
                else:
                    raise HazelcastInstanceNotActiveError()
            return serializer.read(inp)
        except:
            handle_exception(sys.exc_info()[1], sys.exc_info()[2])

    def _calculate_partitioning_hash(self, obj, partitioning_strategy):
        partitioning_hash = 0
        _ps = (
            partitioning_strategy
            if partitioning_strategy is not None
            else self._global_partition_strategy
        )
        pk = _ps(obj)
        if pk is not None and pk is not obj:
            partitioning_key = self.to_data(pk, empty_partitioning_strategy)
            partitioning_hash = (
                0 if partitioning_key is None else partitioning_key.get_partition_hash()
            )
        return partitioning_hash

    def _create_data_output(self):
        return _ObjectDataOutput(self._output_buffer_size, self, self._is_big_endian)

    def _create_data_input(self, data):
        return _ObjectDataInput(data._buffer, DATA_OFFSET, self, self._is_big_endian)

    def destroy(self):
        self._active = False
        self._registry.destroy()


class SerializerRegistry(object):
    def __init__(self, int_type):
        self._global_serializer = None
        self._portable_serializer = None
        self._data_serializer = None
        self._null_serializer = NoneSerializer()
        self._python_serializer = PythonObjectSerializer()

        self._constant_type_ids = {}  # dict of id:serializer
        self._constant_type_dict = {}  # dict of class:serializer

        self._id_dic = {}  # dict of type_id:serializer
        self._type_dict = {}  # dict of class:serializer

        self._registration_lock = RLock()
        self.int_type = int_type
        self._int_type_id = _int_type_to_type_id.get(int_type, None)

    def serializer_by_type_id(self, type_id):
        """Find and return the serializer for the type-id

        Args:
            type_id (int): Type id of the serializer

        Returns:
          The serializer
        """
        if type_id <= 0:
            indx = index_for_default_type(type_id)
            serializer = self._constant_type_ids.get(indx, None)
            if serializer is not None:
                return serializer
        return self._id_dic.get(type_id, None)

    def serializer_for(self, obj):
        """Searches for a serializer for the provided object

        Serializers will be  searched in this order;

        - NULL serializer
        - Default serializers, like primitives, arrays, string and some default types
        - Custom registered types by user
        - Global serializer if registered by user
        - pickle serialization as a fallback

        Args:
            obj: Input object

        Returns:
            The serializer
        """
        # 1-NULL serializer
        if obj is None:
            return self._null_serializer

        obj_type = type(obj)

        # 2-Default serializers, DataSerializable, Portable, primitives, arrays, String, UUID
        # and some helper types(BigInteger etc)
        serializer = self.lookup_default_serializer(obj_type, obj)

        # 3-Custom registered types by user
        if serializer is None:
            serializer = self.lookup_custom_serializer(obj_type)

        # 5-Global serializer if registered by user
        if serializer is None:
            serializer = self.lookup_global_serializer(obj_type)

        # 4 Internal serializer
        if serializer is None:
            serializer = self.lookup_python_serializer(obj_type)

        if serializer is None:
            raise HazelcastSerializationError(
                "There is no suitable serializer for:" + str(obj_type)
            )
        return serializer

    def lookup_default_serializer(self, obj_type, obj):
        if isinstance(obj, IdentifiedDataSerializable):
            return self._data_serializer

        if isinstance(obj, Portable):
            return self._portable_serializer

        if isinstance(obj, six.string_types):
            return self.serializer_by_type_id(CONSTANT_TYPE_STRING)

        # LOCATE NUMERIC TYPES
        if obj_type in six.integer_types:
            type_id = self._int_type_id
            if type_id is None:
                # VAR size
                if MIN_BYTE <= obj <= MAX_BYTE:
                    type_id = CONSTANT_TYPE_BYTE
                elif MIN_SHORT <= obj <= MAX_SHORT:
                    type_id = CONSTANT_TYPE_SHORT
                elif MIN_INT <= obj <= MAX_INT:
                    type_id = CONSTANT_TYPE_INTEGER
                elif MIN_LONG <= obj <= MAX_LONG:
                    type_id = CONSTANT_TYPE_LONG
                else:
                    type_id = JAVA_DEFAULT_TYPE_BIG_INTEGER

            return self.serializer_by_type_id(type_id)

        return self._constant_type_dict.get(obj_type, None)

    def lookup_custom_serializer(self, obj_type):
        serializer = self._type_dict.get(obj_type, None)
        if serializer is not None:
            return serializer
        for super_type in obj_type.__subclasses__():
            serializer = self.register_from_super_type(obj_type, super_type)
            if serializer is not None:
                return serializer
        return None

    def lookup_python_serializer(self, obj_type):
        self.safe_register_serializer(self._python_serializer, obj_type)
        return self._python_serializer

    def lookup_global_serializer(self, obj_type):
        serializer = self._global_serializer
        if serializer is not None:
            self.safe_register_serializer(serializer, obj_type)
        return serializer

    def register_constant_serializer(self, serializer, object_type=None):
        stream_serializer = serializer
        self._constant_type_ids[
            index_for_default_type(stream_serializer.get_type_id())
        ] = stream_serializer
        if object_type is not None:
            self._constant_type_dict[object_type] = stream_serializer

    def safe_register_serializer(self, stream_serializer, obj_type=None):
        with self._registration_lock:
            if obj_type is not None:
                if obj_type in self._constant_type_dict:
                    raise ValueError("[%s] serializer cannot be overridden!" % obj_type)
                current = self._type_dict.get(obj_type, None)
                if current is not None and current.__class__ != stream_serializer.__class__:
                    raise ValueError(
                        "Serializer[%s] has been already registered for type: %s"
                        % (current.__class__, obj_type)
                    )
                else:
                    self._type_dict[obj_type] = stream_serializer
            serializer_type_id = stream_serializer.get_type_id()
            current = self._id_dic.get(serializer_type_id, None)
            if current is not None and current.__class__ != stream_serializer.__class__:
                raise ValueError(
                    "Serializer[%s] has been already registered for type-id: %s"
                    % (current.__class__, serializer_type_id)
                )
            else:
                self._id_dic[serializer_type_id] = stream_serializer
            return current is None

    def register_from_super_type(self, obj_type, super_type):
        serializer = self._type_dict.get(super_type, None)
        if serializer is not None:
            self.safe_register_serializer(serializer, obj_type)
        return serializer

    def destroy(self):
        for serializer in list(self._type_dict.values()):
            serializer.destroy()
        for serializer in list(self._constant_type_dict.values()):
            serializer.destroy()
        self._type_dict.clear()
        self._id_dic.clear()
        self._global_serializer = None
        self._constant_type_dict.clear()



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.protocol import EndpointQualifier
from hazelcast.protocol.builtin import StringCodec

_TYPE_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_TYPE_DECODE_OFFSET = 0
_INITIAL_FRAME_SIZE = _TYPE_ENCODE_OFFSET + INT_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class EndpointQualifierCodec(object):
    @staticmethod
    def encode(buf, endpoint_qualifier, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_int(initial_frame_buf, _TYPE_ENCODE_OFFSET, endpoint_qualifier.type)
        buf.extend(initial_frame_buf)
        CodecUtil.encode_nullable(buf, endpoint_qualifier.identifier, StringCodec.encode)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        type = FixSizedTypesCodec.decode_int(initial_frame.buf, _TYPE_DECODE_OFFSET)
        identifier = CodecUtil.decode_nullable(msg, StringCodec.decode)
        CodecUtil.fast_forward_to_end_frame(msg)
        return EndpointQualifier(type, identifier)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec

# hex: 0x1F0300
_REQUEST_MESSAGE_TYPE = 2032384
# hex: 0x1F0301
_RESPONSE_MESSAGE_TYPE = 2032385

_REQUEST_SESSION_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_SESSION_ID_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(group_id, session_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_SESSION_ID_OFFSET, session_id)
    RaftGroupIdCodec.encode(buf, group_id, True)
    return OutboundMessage(buf, True)



from hazelcast.protocol.codec import (
    multi_map_add_entry_listener_codec,
    multi_map_add_entry_listener_to_key_codec,
    multi_map_clear_codec,
    multi_map_contains_entry_codec,
    multi_map_contains_key_codec,
    multi_map_contains_value_codec,
    multi_map_entry_set_codec,
    multi_map_force_unlock_codec,
    multi_map_get_codec,
    multi_map_is_locked_codec,
    multi_map_key_set_codec,
    multi_map_lock_codec,
    multi_map_put_codec,
    multi_map_remove_codec,
    multi_map_remove_entry_codec,
    multi_map_remove_entry_listener_codec,
    multi_map_size_codec,
    multi_map_try_lock_codec,
    multi_map_unlock_codec,
    multi_map_value_count_codec,
    multi_map_values_codec,
)
from hazelcast.proxy.base import Proxy, EntryEvent, EntryEventType
from hazelcast.util import check_not_none, thread_id, to_millis, ImmutableLazyDataList


class MultiMap(Proxy):
    """A specialized map whose keys can be associated with multiple values."""

    def __init__(self, service_name, name, context):
        super(MultiMap, self).__init__(service_name, name, context)
        self._reference_id_generator = context.lock_reference_id_generator

    def add_entry_listener(
        self, include_value=False, key=None, added_func=None, removed_func=None, clear_all_func=None
    ):
        """Adds an entry listener for this multimap.

        The listener will be notified for all multimap add/remove/clear-all events.

        Args:
            include_value (bool): Whether received event should include the value or not.
            key: Key for filtering the events.
            added_func (function): Function to be called when an entry is added to map.
            removed_func (function): Function to be called when an entry is removed from map.
            clear_all_func (function): Function to be called when entries are cleared from map.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        if key:
            codec = multi_map_add_entry_listener_to_key_codec
            key_data = self._to_data(key)
            request = codec.encode_request(self.name, key_data, include_value, False)
        else:
            codec = multi_map_add_entry_listener_codec
            request = codec.encode_request(self.name, include_value, False)

        def handle_event_entry(
            key, value, old_value, merging_value, event_type, uuid, number_of_affected_entries
        ):
            event = EntryEvent(
                self._to_object,
                key,
                value,
                old_value,
                merging_value,
                event_type,
                uuid,
                number_of_affected_entries,
            )
            if event.event_type == EntryEventType.ADDED and added_func:
                added_func(event)
            elif event.event_type == EntryEventType.REMOVED and removed_func:
                removed_func(event)
            elif event.event_type == EntryEventType.CLEAR_ALL and clear_all_func:
                clear_all_func(event)

        return self._register_listener(
            request,
            lambda r: codec.decode_response(r),
            lambda reg_id: multi_map_remove_entry_listener_codec.encode_request(self.name, reg_id),
            lambda m: codec.handle(m, handle_event_entry),
        )

    def contains_key(self, key):
        """Determines whether this multimap contains an entry with the key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this multimap contains an entry for the specified key,
            ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)

        request = multi_map_contains_key_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, multi_map_contains_key_codec.decode_response)

    def contains_value(self, value):
        """Determines whether this map contains one or more keys for the specified value.

        Args:
            value: The specified value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this multimap contains an entry for the specified value,
            ``False`` otherwise.
        """
        check_not_none(value, "value can't be None")
        value_data = self._to_data(value)
        request = multi_map_contains_value_codec.encode_request(self.name, value_data)
        return self._invoke(request, multi_map_contains_value_codec.decode_response)

    def contains_entry(self, key, value):
        """Returns whether the multimap contains an entry with the value.

        Args:
            key: The specified key.
            value: The specified value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this multimap contains the key-value tuple, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)

        request = multi_map_contains_entry_codec.encode_request(
            self.name, key_data, value_data, thread_id()
        )
        return self._invoke_on_key(
            request, key_data, multi_map_contains_entry_codec.decode_response
        )

    def clear(self):
        """Clears the multimap. Removes all key-value tuples.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = multi_map_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def entry_set(self):
        """Returns the list of key-value tuples in the multimap.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.

        Returns:
            hazelcast.future.Future[list]: The list of key-value tuples in the multimap.
        """

        def handler(message):
            return ImmutableLazyDataList(
                multi_map_entry_set_codec.decode_response(message), self._to_object
            )

        request = multi_map_entry_set_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def get(self, key):
        """Returns the list of values associated with the key. ``None`` if this map does not contain this key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` of the binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in the key's class.

        Warning:
            The list is NOT backed by the multimap, so changes to the map are list reflected in the collection, and
            vice-versa.

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[list]: The list of the values associated with the specified key.
        """
        check_not_none(key, "key can't be None")

        def handler(message):
            return ImmutableLazyDataList(
                multi_map_get_codec.decode_response(message), self._to_object
            )

        key_data = self._to_data(key)
        request = multi_map_get_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def is_locked(self, key):
        """Checks the lock for the specified key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key that is checked for lock.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if lock is acquired, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)

        request = multi_map_is_locked_codec.encode_request(self.name, key_data)
        return self._invoke_on_key(request, key_data, multi_map_is_locked_codec.decode_response)

    def force_unlock(self, key):
        """Releases the lock for the specified key regardless of the lock owner.

        It always successfully unlocks the key, never blocks, and returns immediately.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        request = multi_map_force_unlock_codec.encode_request(
            self.name, key_data, self._reference_id_generator.get_and_increment()
        )
        return self._invoke_on_key(request, key_data)

    def key_set(self):
        """Returns the list of keys in the multimap.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.

        Returns:
            hazelcast.future.Future[list]: A list of the clone of the keys.
        """

        def handler(message):
            return ImmutableLazyDataList(
                multi_map_key_set_codec.decode_response(message), self._to_object
            )

        request = multi_map_key_set_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def lock(self, key, lease_time=None):
        """Acquires the lock for the specified key infinitely or for the specified lease time if provided.

        If the lock is not available, the current thread becomes disabled for thread scheduling purposes and lies
        dormant until the lock has been acquired.

        Scope of the lock is this map only. Acquired lock is only for the key in this map.

        Locks are re-entrant; so, if the key is locked N times, it should be unlocked N times before another thread can
        acquire it.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.
            lease_time (int): Time in seconds to wait before releasing the lock.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        request = multi_map_lock_codec.encode_request(
            self.name,
            key_data,
            thread_id(),
            to_millis(lease_time),
            self._reference_id_generator.get_and_increment(),
        )
        return self._invoke_on_key(request, key_data)

    def remove(self, key, value):
        """Removes the given key-value tuple from the multimap.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key of the entry to remove.
            value: The value of the entry to remove.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the size of the multimap changed after the remove operation,
            ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(key, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = multi_map_remove_entry_codec.encode_request(
            self.name, key_data, value_data, thread_id()
        )
        return self._invoke_on_key(request, key_data, multi_map_remove_entry_codec.decode_response)

    def remove_all(self, key):
        """Removes all the entries with the given key and returns the value list associated with this key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Warning:
            The returned list is NOT backed by the map, so changes to the map are NOT reflected in the list, and
            vice-versa.

        Args:
            key: The key of the entries to remove.

        Returns:
            hazelcast.future.Future[list]: The collection of removed values associated with the given key.
        """
        check_not_none(key, "key can't be None")

        def handler(message):
            return ImmutableLazyDataList(
                multi_map_remove_codec.decode_response(message), self._to_object
            )

        key_data = self._to_data(key)
        request = multi_map_remove_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def put(self, key, value):
        """Stores a key-value tuple in the multimap.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key to be stored.
            value: The value to be stored.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if size of the multimap is increased,
            ``False`` if the multimap already contains the key-value tuple.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        request = multi_map_put_codec.encode_request(self.name, key_data, value_data, thread_id())
        return self._invoke_on_key(request, key_data, multi_map_put_codec.decode_response)

    def remove_entry_listener(self, registration_id):
        """Removes the specified entry listener.

        Returns silently if there is no such listener added before.

        Args:
            registration_id (str): Id of registered listener.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if registration is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def size(self):
        """Returns the number of entries in this multimap.

        Returns:
            hazelcast.future.Future[int]: Number of entries in this multimap.
        """
        request = multi_map_size_codec.encode_request(self.name)
        return self._invoke(request, multi_map_size_codec.decode_response)

    def value_count(self, key):
        """Returns the number of values that match the given key in the multimap.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key whose values count is to be returned.

        Returns:
            hazelcast.future.Future[int]: The number of values that match the given key in the multimap.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        request = multi_map_value_count_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, multi_map_value_count_codec.decode_response)

    def values(self):
        """Returns the list of values in the multimap.

        Warning:
            The returned list is NOT backed by the map, so changes to the map are NOT reflected in the list, and
            vice-versa.

        Returns:
            hazelcast.future.Future[list]: The list of values in the multimap.
        """

        def handler(message):
            return ImmutableLazyDataList(
                multi_map_values_codec.decode_response(message), self._to_object
            )

        request = multi_map_values_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def try_lock(self, key, lease_time=None, timeout=0):
        """Tries to acquire the lock for the specified key.

        When the lock is not available:

        - If the timeout is not provided, the current thread doesn't wait and returns ``False`` immediately.
        - If the timeout is provided, the current thread becomes disabled for thread scheduling purposes and lies
          dormant until one of the followings happens:

            - The lock is acquired by the current thread, or
            - The specified waiting time elapses.

        If the lease time is provided, lock will be released after this time elapses.

        Args:
            key: Key to lock in this map.
            lease_time (int): Time in seconds to wait before releasing the lock.
            timeout (int): Maximum time in seconds to wait for the lock.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the lock was acquired, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        request = multi_map_try_lock_codec.encode_request(
            self.name,
            key_data,
            thread_id(),
            to_millis(lease_time),
            to_millis(timeout),
            self._reference_id_generator.get_and_increment(),
        )
        return self._invoke_on_key(request, key_data, multi_map_try_lock_codec.decode_response)

    def unlock(self, key):
        """Releases the lock for the specified key. It never blocks and returns immediately.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        request = multi_map_unlock_codec.encode_request(
            self.name, key_data, thread_id(), self._reference_id_generator.get_and_increment()
        )
        return self._invoke_on_key(request, key_data)



"""Hazelcast Core objects and constants."""
import json

from hazelcast import six
from hazelcast.six.moves import range
from hazelcast import util

CLIENT_TYPE = "PYH"
SERIALIZATION_VERSION = 1


class MemberInfo(object):
    """
    Represents a member in the cluster with its address, uuid, lite member status, attributes and version.
    """

    __slots__ = ("address", "uuid", "attributes", "lite_member", "version")

    def __init__(self, address, uuid, attributes, lite_member, version, *_):
        self.address = address
        """
        hazelcast.core.Address: Address of the member.
        """

        self.uuid = uuid
        """
        uuid.UUID: UUID of the member.
        """

        self.attributes = attributes
        """
        dict[str, str]: Configured attributes of the member.
        """

        self.lite_member = lite_member
        """
        bool: ``True`` if the member is a lite member, ``False`` otherwise.
        Lite members do not own any partition.
        """

        self.version = version
        """
        hazelcast.core.MemberVersion: Hazelcast codebase version of the member.
        """

    def __str__(self):
        return "Member [%s]:%s - %s" % (self.address.host, self.address.port, self.uuid)

    def __repr__(self):
        return "Member(address=%s, uuid=%s, attributes=%s, lite_member=%s, version=%s)" % (
            self.address,
            self.uuid,
            self.attributes,
            self.lite_member,
            self.version,
        )

    def __hash__(self):
        return hash((self.address, self.uuid))

    def __eq__(self, other):
        return (
            isinstance(other, MemberInfo)
            and self.address == other.address
            and self.uuid == other.uuid
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class Address(object):
    """Represents an address of a member in the cluster."""

    def __init__(self, host, port):
        self.host = host
        """
        str: Host of the address.
        """

        self.port = port
        """
        int: Port of the address.
        """

    def __repr__(self):
        return "Address(host=%s, port=%d)" % (self.host, self.port)

    def __hash__(self):
        return hash((self.host, self.port))

    def __eq__(self, other):
        return isinstance(other, Address) and self.host == other.host and self.port == other.port

    def __ne__(self, other):
        return not self.__eq__(other)


class AddressHelper(object):
    @staticmethod
    def get_possible_addresses(address):
        address = AddressHelper.address_from_str(address)
        possible_port = address.port
        port_try_count = 1
        if possible_port == -1:
            port_try_count = 3
            possible_port = 5701

        addresses = []
        for i in range(port_try_count):
            addresses.append(Address(address.host, possible_port + i))

        # primary, secondary
        return [addresses.pop(0)], addresses

    @staticmethod
    def address_from_str(address, port=-1):
        bracket_start_idx = address.find("[")
        bracket_end_idx = address.find("]", bracket_start_idx)
        colon_idx = address.find(":")
        last_colon_idx = address.rfind(":")

        if -1 < colon_idx < last_colon_idx:
            # IPv6
            if bracket_start_idx == 0 and bracket_end_idx > bracket_start_idx:
                host = address[bracket_start_idx + 1 : bracket_end_idx]
                if last_colon_idx == (bracket_end_idx + 1):
                    port = int(address[last_colon_idx + 1 :])
            else:
                host = address
        elif colon_idx > 0 and colon_idx == last_colon_idx:
            host = address[:colon_idx]
            port = int(address[colon_idx + 1 :])
        else:
            host = address
        return Address(host, port)


class DistributedObjectInfo(object):
    def __init__(self, service_name, name):
        self.service_name = service_name
        self.name = name

    def __repr__(self):
        return "DistributedObjectInfo(serviceName=%s, name=%s)" % (self.service_name, self.name)

    def __hash__(self):
        return hash((self.name, self.service_name))

    def __eq__(self, other):
        if isinstance(other, DistributedObjectInfo):
            return self.name == other.name and self.service_name == other.service_name
        return False


class DistributedObjectEventType(object):
    """Type of the distributed object event."""

    CREATED = "CREATED"
    """
    DistributedObject is created.
    """

    DESTROYED = "DESTROYED"
    """
    DistributedObject is destroyed.
    """


class DistributedObjectEvent(object):
    """Distributed Object Event"""

    def __init__(self, name, service_name, event_type, source):
        self.name = name
        """
        str: Name of the distributed object.
        """

        self.service_name = service_name
        """
        str: Service name of the distributed object.
        """

        self.event_type = event_type
        """
        str: Event type. Either ``CREATED`` or ``DESTROYED``.
        """

        self.source = source
        """
        uuid.UUID: UUID of the member that fired the event.
        """

    def __repr__(self):
        return "DistributedObjectEvent(name=%s, service_name=%s, event_type=%s, source=%s)" % (
            self.name,
            self.service_name,
            self.event_type,
            self.source,
        )


class SimpleEntryView(object):
    """EntryView represents a readonly view of a map entry."""

    def __init__(
        self,
        key,
        value,
        cost,
        creation_time,
        expiration_time,
        hits,
        last_access_time,
        last_stored_time,
        last_update_time,
        version,
        ttl,
        max_idle,
    ):
        self.key = key
        """
        The key of the entry.
        """

        self.value = value
        """
        The value of the entry.
        """

        self.cost = cost
        """
        int: The cost in bytes of the entry.
        """

        self.creation_time = creation_time
        """
        int: The creation time of the entry.
        """

        self.expiration_time = expiration_time
        """
        int: The expiration time of the entry.
        """

        self.hits = hits
        """
        int: Number of hits of the entry.
        """

        self.last_access_time = last_access_time
        """
        int: The last access time for the entry.
        """

        self.last_stored_time = last_stored_time
        """
        int: The last store time for the value.
        """

        self.last_update_time = last_update_time
        """
        int: The last time the value was updated.
        """

        self.version = version
        """
        int: The version of the entry.
        """

        self.ttl = ttl
        """
        int: The last set time to live milliseconds.
        """

        self.max_idle = max_idle
        """
        int: The last set max idle time in milliseconds.
        """

    def __repr__(self):
        return (
            "SimpleEntryView(key=%s, value=%s, cost=%s, creation_time=%s, "
            "expiration_time=%s, hits=%s, last_access_time=%s, last_stored_time=%s, "
            "last_update_time=%s, version=%s, ttl=%s, max_idle=%s"
            % (
                self.key,
                self.value,
                self.cost,
                self.creation_time,
                self.expiration_time,
                self.hits,
                self.last_access_time,
                self.last_stored_time,
                self.last_update_time,
                self.version,
                self.ttl,
                self.max_idle,
            )
        )


class HazelcastJsonValue(object):
    """HazelcastJsonValue is a wrapper for JSON formatted strings.

    It is preferred to store HazelcastJsonValue instead of Strings for JSON formatted strings.
    Users can run predicates and use indexes on the attributes of the underlying
    JSON strings.

    HazelcastJsonValue is queried using Hazelcast's querying language.
    See `Distributed Query section <https://github.com/hazelcast/hazelcast-python-client#77-distributed-query>`_.

    In terms of querying, numbers in JSON strings are treated as either
    Long or Double in the Java side. str, bool and None
    are treated as String, boolean and null respectively.

    HazelcastJsonValue keeps given string as it is. Strings are not
    checked for being valid. Ill-formatted JSON strings may cause false
    positive or false negative results in queries.

    HazelcastJsonValue can also be constructed from JSON serializable objects.
    In that case, objects are converted to JSON strings and stored as such.
    If an error occurs during the conversion, it is raised directly.

    None values are not allowed.
    """

    def __init__(self, value):
        util.check_not_none(value, "JSON string or the object cannot be None.")
        if isinstance(value, six.string_types):
            self._json_string = value
        else:
            self._json_string = json.dumps(value)

    def to_string(self):
        """Returns unaltered string that was used to create this object.

        Returns:
            str: The original string.
        """
        return self._json_string

    def loads(self):
        """Deserializes the string that was used to create this object
        and returns as Python object.

        Returns:
            any: The Python object represented by the original string.
        """
        return json.loads(self._json_string)

    def __eq__(self, other):
        return isinstance(other, HazelcastJsonValue) and self._json_string == other._json_string

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._json_string)

    def __repr__(self):
        return self._json_string


class MemberVersion(object):
    """
    Determines the Hazelcast codebase version in terms of major.minor.patch version.
    """

    __slots__ = ("major", "minor", "patch")

    def __init__(self, major, minor, patch):
        self.major = major
        self.minor = minor
        self.patch = patch


class MapEntry(object):
    """
    Represents the entry of a Map, with key and value fields.
    """

    __slots__ = ("_key", "_value")

    def __init__(self, key=None, value=None):
        self._key = key
        self._value = value

    @property
    def key(self):
        """Key of the entry."""
        return self._key

    @property
    def value(self):
        """Value of the entry."""
        return self._value



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.sql_query_id_codec import SqlQueryIdCodec
from hazelcast.protocol.builtin import SqlPageCodec
from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.codec.custom.sql_error_codec import SqlErrorCodec

# hex: 0x210500
_REQUEST_MESSAGE_TYPE = 2163968
# hex: 0x210501
_RESPONSE_MESSAGE_TYPE = 2163969

_REQUEST_CURSOR_BUFFER_SIZE_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_CURSOR_BUFFER_SIZE_OFFSET + INT_SIZE_IN_BYTES


def encode_request(query_id, cursor_buffer_size):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_CURSOR_BUFFER_SIZE_OFFSET, cursor_buffer_size)
    SqlQueryIdCodec.encode(buf, query_id, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    msg.next_frame()
    response = dict()
    response["row_page"] = CodecUtil.decode_nullable(msg, SqlPageCodec.decode)
    response["error"] = CodecUtil.decode_nullable(msg, SqlErrorCodec.decode)
    return response



import struct

"""
Constants
"""
BYTE_SIZE_IN_BYTES = 1
BOOLEAN_SIZE_IN_BYTES = 1
SHORT_SIZE_IN_BYTES = 2
CHAR_SIZE_IN_BYTES = 2
INT_SIZE_IN_BYTES = 4
FLOAT_SIZE_IN_BYTES = 4
LONG_SIZE_IN_BYTES = 8
DOUBLE_SIZE_IN_BYTES = 8
UUID_SIZE_IN_BYTES = 17  # bool + long + long

LE_INT = struct.Struct("<i")
LE_UINT = struct.Struct("<I")
LE_INT8 = struct.Struct("<b")
LE_UINT8 = struct.Struct("<B")
LE_INT16 = struct.Struct("<h")
LE_UINT16 = struct.Struct("<H")
LE_LONG = struct.Struct("<q")
LE_ULONG = struct.Struct("<Q")
LE_FLOAT = struct.Struct("<f")
LE_DOUBLE = struct.Struct("<d")

BE_INT = struct.Struct(">i")
BE_INT8 = struct.Struct(">b")
BE_UINT8 = struct.Struct(">B")
BE_INT16 = struct.Struct(">h")
BE_UINT16 = struct.Struct(">H")
BE_LONG = struct.Struct(">q")
BE_FLOAT = struct.Struct(">f")
BE_DOUBLE = struct.Struct(">d")

BIG_ENDIAN = 2
LITTLE_ENDIAN = 1

NULL_ARRAY_LENGTH = -1

# LIMITS
MAX_BYTE = 2 ** 7 - 1
MIN_BYTE = -(2 ** 7)

MAX_SHORT = 2 ** 15 - 1
MIN_SHORT = -(2 ** 15)

MAX_INT = 2 ** 31 - 1
MIN_INT = -(2 ** 31)

MAX_LONG = 2 ** 63 - 1
MIN_LONG = -(2 ** 63)

UUID_MSB_SHIFT = 64
UUID_MSB_MASK = 0xFFFFFFFFFFFFFFFF << UUID_MSB_SHIFT
UUID_LSB_MASK = 0xFFFFFFFFFFFFFFFF



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.codec.custom.simple_entry_view_codec import SimpleEntryViewCodec
from hazelcast.protocol.builtin import CodecUtil

# hex: 0x011D00
_REQUEST_MESSAGE_TYPE = 72960
# hex: 0x011D01
_RESPONSE_MESSAGE_TYPE = 72961

_REQUEST_THREAD_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES
_RESPONSE_MAX_IDLE_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(name, key, thread_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, key, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["max_idle"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_MAX_IDLE_OFFSET)
    response["response"] = CodecUtil.decode_nullable(msg, SimpleEntryViewCodec.decode)
    return response



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.sql import _SqlError
from hazelcast.protocol.builtin import StringCodec

_CODE_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_CODE_DECODE_OFFSET = 0
_ORIGINATING_MEMBER_ID_ENCODE_OFFSET = _CODE_ENCODE_OFFSET + INT_SIZE_IN_BYTES
_ORIGINATING_MEMBER_ID_DECODE_OFFSET = _CODE_DECODE_OFFSET + INT_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _ORIGINATING_MEMBER_ID_ENCODE_OFFSET + UUID_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class SqlErrorCodec(object):
    @staticmethod
    def encode(buf, sql_error, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_int(initial_frame_buf, _CODE_ENCODE_OFFSET, sql_error.code)
        FixSizedTypesCodec.encode_uuid(initial_frame_buf, _ORIGINATING_MEMBER_ID_ENCODE_OFFSET, sql_error.originating_member_id)
        buf.extend(initial_frame_buf)
        CodecUtil.encode_nullable(buf, sql_error.message, StringCodec.encode)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        code = FixSizedTypesCodec.decode_int(initial_frame.buf, _CODE_DECODE_OFFSET)
        originating_member_id = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _ORIGINATING_MEMBER_ID_DECODE_OFFSET)
        message = CodecUtil.decode_nullable(msg, StringCodec.decode)
        CodecUtil.fast_forward_to_end_frame(msg)
        return _SqlError(code, message, originating_member_id)



from hazelcast.core import Address, MapEntry
from hazelcast.serialization.api import IdentifiedDataSerializable
from hazelcast.serialization.data import Data
from hazelcast.util import to_millis


# Had to put this class into the serialization module
# since it is required in two places.
# 1 - In the serialization service construction, since
#     its factory is registered.
# 2 - In ReliableTopic proxy, since publish wraps messages
#     into this type.
# If this class was at the proxy module, we would get
# cyclic dependencies.
class ReliableTopicMessage(IdentifiedDataSerializable):
    """The Object that is going to be stored in the Ringbuffer.
    It contains the actual message payload and some metadata.
    """

    FACTORY_ID = -9
    CLASS_ID = 2

    def __init__(self, publish_time=None, publisher_address=None, payload=None):
        # publish_time is in seconds but server sends/expects to receive
        # it in milliseconds.
        self.publish_time = publish_time
        self.publisher_address = publisher_address
        self.payload = payload

    def read_data(self, object_data_input):
        self.publish_time = object_data_input.read_long() / 1000.0
        self.publisher_address = object_data_input.read_object()
        self.payload = _read_data_from(object_data_input)

    def write_data(self, object_data_output):
        object_data_output.write_long(to_millis(self.publish_time))
        object_data_output.write_object(self.publisher_address)
        _write_data_to(object_data_output, self.payload)

    def get_factory_id(self):
        return self.FACTORY_ID

    def get_class_id(self):
        return self.CLASS_ID


# Inherits from the Address, as we cannot implement
# IDS in the core module due to cyclic import problems.
# This was needed to construct the ReliableTopic messages
# sent from the server as they have a non-None address
# field.
class IdentifiedAddress(Address, IdentifiedDataSerializable):

    FACTORY_ID = 0
    CLASS_ID = 1

    def __init__(self, host=None, port=None):
        super(IdentifiedAddress, self).__init__(host, port)

    def write_data(self, object_data_output):
        # skip writing it as we don't need it
        pass

    def read_data(self, object_data_input):
        self.port = object_data_input.read_int()
        # skip type as it is unused in the Python client
        object_data_input.read_byte()
        self.host = object_data_input.read_string()

    def get_factory_id(self):
        return self.FACTORY_ID

    def get_class_id(self):
        return self.CLASS_ID


# Values are canonicalized in the server-side. No need to
# do the same here.
class CanonicalizingHashSet(set, IdentifiedDataSerializable):

    FACTORY_ID = -29
    CLASS_ID = 19

    def write_data(self, object_data_output):
        pass

    def read_data(self, object_data_input):
        count = object_data_input.read_int()
        for _ in range(count):
            self.add(object_data_input.read_object())

    def get_factory_id(self):
        return self.FACTORY_ID

    def get_class_id(self):
        return self.CLASS_ID


class IdentifiedMapEntry(MapEntry, IdentifiedDataSerializable):

    FACTORY_ID = -4
    CLASS_ID = 120

    def write_data(self, object_data_output):
        pass

    def read_data(self, object_data_input):
        self._key = object_data_input.read_object()
        self._value = object_data_input.read_object()

    def get_factory_id(self):
        return self.FACTORY_ID

    def get_class_id(self):
        return self.CLASS_ID


def _read_data_from(inp):
    array = inp.read_byte_array()
    if array is None:
        return None
    return Data(array)


def _write_data_to(out, data):
    out.write_byte_array(data.to_bytes())



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil

# hex: 0x0A0500
_REQUEST_MESSAGE_TYPE = 656640
# hex: 0x0A0501
_RESPONSE_MESSAGE_TYPE = 656641

_REQUEST_RETURN_OLD_VALUE_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_RETURN_OLD_VALUE_OFFSET + BOOLEAN_SIZE_IN_BYTES


def encode_request(group_id, name, new_value, return_old_value):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_RETURN_OLD_VALUE_OFFSET, return_old_value)
    RaftGroupIdCodec.encode(buf, group_id)
    StringCodec.encode(buf, name)
    CodecUtil.encode_nullable(buf, new_value, DataCodec.encode, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    msg.next_frame()
    return CodecUtil.decode_nullable(msg, DataCodec.decode)



from hazelcast.errors import HazelcastSerializationError
from hazelcast.serialization import bits
from hazelcast.serialization.api import PortableReader
from hazelcast.serialization.portable.classdef import FieldType
from hazelcast.six.moves import range


class DefaultPortableReader(PortableReader):
    def __init__(self, portable_serializer, data_input, class_def):
        self._portable_serializer = portable_serializer
        self._in = data_input
        self._class_def = class_def
        try:
            # final position after portable is read
            self._final_pos = data_input.read_int()
            # field count
            field_count = data_input.read_int()
        except Exception:
            raise HazelcastSerializationError()
        if field_count != class_def.get_field_count():
            raise ValueError(
                "Field count(%s) in stream does not match! %s" % (field_count, class_def)
            )
        self._offset = data_input.position()
        self._raw = False

    def get_version(self):
        return self._class_def.version

    def has_field(self, field_name):
        return self._class_def.has_field(field_name)

    def get_field_names(self):
        return self._class_def.get_field_names()

    def get_field_type(self, field_name):
        return self._class_def.get_field_type(field_name)

    def get_field_class_id(self, field_name):
        return self._class_def.get_field_class_id(field_name)

    def read_boolean(self, field_name):
        pos = self._read_position(field_name, FieldType.BOOLEAN)
        return self._in.read_boolean(pos)

    def read_byte(self, field_name):
        pos = self._read_position(field_name, FieldType.BYTE)
        return self._in.read_byte(pos)

    def read_char(self, field_name):
        pos = self._read_position(field_name, FieldType.CHAR)
        return self._in.read_char(pos)

    def read_short(self, field_name):
        pos = self._read_position(field_name, FieldType.SHORT)
        return self._in.read_short(pos)

    def read_int(self, field_name):
        pos = self._read_position(field_name, FieldType.INT)
        return self._in.read_int(pos)

    def read_long(self, field_name):
        pos = self._read_position(field_name, FieldType.LONG)
        return self._in.read_long(pos)

    def read_float(self, field_name):
        pos = self._read_position(field_name, FieldType.FLOAT)
        return self._in.read_float(pos)

    def read_double(self, field_name):
        pos = self._read_position(field_name, FieldType.DOUBLE)
        return self._in.read_double(pos)

    def read_string(self, field_name):
        cur_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.STRING)
            self._in.set_position(pos)
            return self._in.read_string()
        finally:
            self._in.set_position(cur_pos)

    def read_portable(self, field_name):
        cur_pos = self._in.position()
        try:
            fd = self._class_def.get_field(field_name)
            if fd is None:
                raise self._create_unknown_field_exception(field_name)
            if fd.field_type != FieldType.PORTABLE:
                raise HazelcastSerializationError("Not a Portable field: %s" % field_name)

            pos = self._read_position_by_field_def(fd)
            self._in.set_position(pos)

            is_none = self._in.read_boolean()
            factory_id = self._in.read_int()
            class_id = self._in.read_int()

            _check_factory_and_class(fd, factory_id, class_id)

            if is_none:
                return None
            return self._portable_serializer.read_internal(self._in, factory_id, class_id)
        finally:
            self._in.set_position(cur_pos)

    def read_boolean_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.BOOLEAN_ARRAY)
            self._in.set_position(pos)
            return self._in.read_boolean_array()
        finally:
            self._in.set_position(current_pos)

    def read_byte_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.BYTE_ARRAY)
            self._in.set_position(pos)
            return self._in.read_byte_array()
        finally:
            self._in.set_position(current_pos)

    def read_char_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.CHAR_ARRAY)
            self._in.set_position(pos)
            return self._in.read_char_array()
        finally:
            self._in.set_position(current_pos)

    def read_short_array(self, field_name):
        pass
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.SHORT_ARRAY)
            self._in.set_position(pos)
            return self._in.read_short_array()
        finally:
            self._in.set_position(current_pos)

    def read_int_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.INT_ARRAY)
            self._in.set_position(pos)
            return self._in.read_int_array()
        finally:
            self._in.set_position(current_pos)

    def read_long_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.LONG_ARRAY)
            self._in.set_position(pos)
            return self._in.read_long_array()
        finally:
            self._in.set_position(current_pos)

    def read_float_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.FLOAT_ARRAY)
            self._in.set_position(pos)
            return self._in.read_float_array()
        finally:
            self._in.set_position(current_pos)

    def read_double_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.DOUBLE_ARRAY)
            self._in.set_position(pos)
            return self._in.read_double_array()
        finally:
            self._in.set_position(current_pos)

    def read_string_array(self, field_name):
        current_pos = self._in.position()
        try:
            pos = self._read_position(field_name, FieldType.STRING_ARRAY)
            self._in.set_position(pos)
            return self._in.read_string_array()
        finally:
            self._in.set_position(current_pos)

    def read_portable_array(self, field_name):
        current_pos = self._in.position()
        try:
            fd = self._class_def.get_field(field_name)
            if fd is None:
                raise self._create_unknown_field_exception(field_name)
            if fd.field_type != FieldType.PORTABLE_ARRAY:
                raise HazelcastSerializationError("Not a portable array field: %s" % field_name)

            pos = self._read_position_by_field_def(fd)
            self._in.set_position(pos)

            length = self._in.read_int()
            factory_id = self._in.read_int()
            class_id = self._in.read_int()
            if length == bits.NULL_ARRAY_LENGTH:
                return None

            _check_factory_and_class(fd, factory_id, class_id)

            portables = [None] * length
            if length > 0:
                offset = self._in.position()
                for i in range(0, length):
                    start = self._in.read_int(offset + i * bits.INT_SIZE_IN_BYTES)
                    self._in.set_position(start)
                    portables[i] = self._portable_serializer.read_internal(
                        self._in, factory_id, class_id
                    )
            return portables
        finally:
            self._in.set_position(current_pos)

    def read_utf(self, field_name):
        return self.read_string(field_name)

    def read_utf_array(self, field_name):
        return self.read_string_array(field_name)

    def get_raw_data_input(self):
        if not self._raw:
            pos = self._in.read_int(
                self._offset + self._class_def.get_field_count() * bits.INT_SIZE_IN_BYTES
            )
            self._in.set_position(pos)
        self._raw = True
        return self._in

    def end(self):
        self._in.set_position(self._final_pos)

    def _read_position(self, field_name, field_type):
        if self._raw:
            raise HazelcastSerializationError(
                "Cannot read Portable fields after get_raw_data_input() is called!"
            )
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return self._read_nested_position(field_name, field_type)
        if fd.field_type != field_type:
            raise HazelcastSerializationError("Not a '%s' field: %s" % (field_type, field_name))
        return self._read_position_by_field_def(fd)

    def _read_nested_position(self, field_name, field_type):
        field_names = field_name.split(".")
        if len(field_names) > 1:
            fd = None
            _reader = self
            for i in range(0, len(field_names)):
                fd = _reader._class_def.get_field(field_names[i])
                if fd is None:
                    break
                if i == len(field_names) - 1:
                    break
                pos = _reader._read_position_by_field_def(fd)
                self._in.set_position(pos)
                is_none = self._in.read_boolean()
                if is_none:
                    raise ValueError("Parent field is null: %s" % field_names[i])
                _reader = self._portable_serializer.create_default_reader(self._in)
            if fd is None:
                raise self._create_unknown_field_exception(field_name)
            if fd.field_type != field_type:
                raise HazelcastSerializationError("Not a '%s' field: %s" % (field_type, field_name))
            return _reader._read_position_by_field_def(fd)
        raise self._create_unknown_field_exception(field_name)

    def _create_unknown_field_exception(self, field_name):
        return HazelcastSerializationError(
            "Unknown field name: '%s' for ClassDefinition(id=%s, version=%s)"
            % (field_name, self._class_def.class_id, self._class_def.version)
        )

    def _read_position_by_field_def(self, fd):
        pos = self._in.read_int(self._offset + fd.index * bits.INT_SIZE_IN_BYTES)
        _len = self._in.read_short(pos)
        # name + len + type
        return pos + bits.SHORT_SIZE_IN_BYTES + _len + 1


def _check_factory_and_class(field_def, factory_id, class_id):
    if factory_id != field_def.factory_id:
        raise ValueError(
            "Invalid factoryId! Expected: %s, Current: %s" % (factory_id, field_def.factory_id)
        )
    if class_id != field_def.class_id:
        raise ValueError(
            "Invalid classId! Expected: %s, Current: %s" % (class_id, field_def.class_id)
        )


class MorphingPortableReader(DefaultPortableReader):
    def read_short(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0
        elif fd.field_type == FieldType.SHORT:
            return super(MorphingPortableReader, self).read_short(field_name)
        elif fd.field_type == FieldType.BYTE:
            return super(MorphingPortableReader, self).read_byte(field_name)
        else:
            raise self.create_incompatible_class_change_error(fd, FieldType.SHORT)

    def read_int(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0
        elif fd.field_type == FieldType.INT:
            return super(MorphingPortableReader, self).read_int(field_name)
        elif fd.field_type == FieldType.BYTE:
            return super(MorphingPortableReader, self).read_byte(field_name)
        elif fd.field_type == FieldType.CHAR:
            return super(MorphingPortableReader, self).read_char(field_name)
        elif fd.field_type == FieldType.SHORT:
            return super(MorphingPortableReader, self).read_short(field_name)
        else:
            raise self.create_incompatible_class_change_error(fd, FieldType.INT)

    def read_long(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0
        elif fd.field_type == FieldType.LONG:
            return super(MorphingPortableReader, self).read_long(field_name)
        elif fd.field_type == FieldType.INT:
            return super(MorphingPortableReader, self).read_int(field_name)
        elif fd.field_type == FieldType.BYTE:
            return super(MorphingPortableReader, self).read_byte(field_name)
        elif fd.field_type == FieldType.CHAR:
            return super(MorphingPortableReader, self).read_char(field_name)
        elif fd.field_type == FieldType.SHORT:
            return super(MorphingPortableReader, self).read_short(field_name)
        else:
            raise self.create_incompatible_class_change_error(fd, FieldType.LONG)

    def read_float(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0
        elif fd.field_type == FieldType.FLOAT:
            return super(MorphingPortableReader, self).read_float(field_name)
        elif fd.field_type == FieldType.INT:
            return super(MorphingPortableReader, self).read_int(field_name)
        elif fd.field_type == FieldType.BYTE:
            return super(MorphingPortableReader, self).read_byte(field_name)
        elif fd.field_type == FieldType.CHAR:
            return super(MorphingPortableReader, self).read_char(field_name)
        elif fd.field_type == FieldType.SHORT:
            return super(MorphingPortableReader, self).read_short(field_name)
        else:
            raise self.create_incompatible_class_change_error(fd, FieldType.FLOAT)

    def read_double(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0.0
        elif fd.field_type == FieldType.DOUBLE:
            return super(MorphingPortableReader, self).read_double(field_name)
        elif fd.field_type == FieldType.LONG:
            return super(MorphingPortableReader, self).read_long(field_name)
        elif fd.field_type == FieldType.FLOAT:
            return super(MorphingPortableReader, self).read_float(field_name)
        elif fd.field_type == FieldType.INT:
            return super(MorphingPortableReader, self).read_int(field_name)
        elif fd.field_type == FieldType.BYTE:
            return super(MorphingPortableReader, self).read_byte(field_name)
        elif fd.field_type == FieldType.CHAR:
            return super(MorphingPortableReader, self).read_char(field_name)
        elif fd.field_type == FieldType.SHORT:
            return super(MorphingPortableReader, self).read_short(field_name)
        else:
            raise self.create_incompatible_class_change_error(fd, FieldType.DOUBLE)

    def read_byte(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0
        self.validate_type_compatibility(fd, FieldType.BYTE)
        return super(MorphingPortableReader, self).read_byte(field_name)

    def read_boolean(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return False
        self.validate_type_compatibility(fd, FieldType.BOOLEAN)
        return super(MorphingPortableReader, self).read_boolean(field_name)

    def read_char(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return 0
        self.validate_type_compatibility(fd, FieldType.CHAR)
        return super(MorphingPortableReader, self).read_char(field_name)

    def read_string(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.STRING)
        return super(MorphingPortableReader, self).read_string(field_name)

    def read_string_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.STRING_ARRAY)
        return super(MorphingPortableReader, self).read_string_array(field_name)

    def read_short_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.SHORT_ARRAY)
        return super(MorphingPortableReader, self).read_short_array(field_name)

    def read_int_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.INT_ARRAY)
        return super(MorphingPortableReader, self).read_int_array(field_name)

    def read_long_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.LONG_ARRAY)
        return super(MorphingPortableReader, self).read_long_array(field_name)

    def read_float_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.FLOAT_ARRAY)
        return super(MorphingPortableReader, self).read_float_array(field_name)

    def read_double_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.DOUBLE_ARRAY)
        return super(MorphingPortableReader, self).read_double_array(field_name)

    def read_char_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.CHAR_ARRAY)
        return super(MorphingPortableReader, self).read_char_array(field_name)

    def read_byte_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.BYTE_ARRAY)
        return super(MorphingPortableReader, self).read_byte_array(field_name)

    def read_boolean_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.BOOLEAN_ARRAY)
        return super(MorphingPortableReader, self).read_boolean_array(field_name)

    def read_portable(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.PORTABLE)
        return super(MorphingPortableReader, self).read_portable(field_name)

    def read_portable_array(self, field_name):
        fd = self._class_def.get_field(field_name)
        if fd is None:
            return None
        self.validate_type_compatibility(fd, FieldType.PORTABLE_ARRAY)
        return super(MorphingPortableReader, self).read_portable_array(field_name)

    def read_utf(self, field_name):
        return self.read_string(field_name)

    def read_utf_array(self, field_name):
        return self.read_string_array(field_name)

    def validate_type_compatibility(self, field_def, expected_type):
        if field_def.field_type != expected_type:
            raise self.create_incompatible_class_change_error(field_def, expected_type)

    def create_incompatible_class_change_error(self, field_def, expected_type):
        return TypeError(
            "Incompatible to read %s from %s while reading field: %s on %s"
            % (expected_type, field_def.field_type, field_def.field_name, self._class_def)
        )



__version__ = "4.2.1"

# Set the default handler to "hazelcast" loggers
# to avoid "No handlers could be found" warnings.
import logging

try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())

from hazelcast.client import HazelcastClient



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.sql_query_id_codec import SqlQueryIdCodec

# hex: 0x210300
_REQUEST_MESSAGE_TYPE = 2163456
# hex: 0x210301
_RESPONSE_MESSAGE_TYPE = 2163457

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(query_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    SqlQueryIdCodec.encode(buf, query_id, True)
    return OutboundMessage(buf, False)



from hazelcast.serialization.api import *
from hazelcast.serialization.bits import *
from hazelcast.six.moves import range


class _ObjectDataOutput(ObjectDataOutput):
    def __init__(self, init_size, serialization_service, is_big_endian=True):
        self._init_size = init_size
        self._buffer = bytearray(init_size)
        self._service = serialization_service
        self._is_big_endian = is_big_endian
        self._pos = 0
        # Local cache struct formats according to endianness
        self._FMT_INT = BE_INT if self._is_big_endian else LE_INT
        self._FMT_SHORT = BE_INT16 if self._is_big_endian else LE_INT16
        self._CHAR_ENCODING = "utf_16_be" if self._is_big_endian else "utf_16_le"
        self._FMT_LONG = BE_LONG if self._is_big_endian else LE_LONG
        self._FMT_FLOAT = BE_FLOAT if self._is_big_endian else LE_FLOAT
        self._FMT_DOUBLE = BE_DOUBLE if self._is_big_endian else LE_DOUBLE

    def _write(self, val):
        self._ensure_available(BYTE_SIZE_IN_BYTES)
        self._buffer[self._pos] = val
        self._pos += BYTE_SIZE_IN_BYTES

    def write_from(self, buff, offset=None, length=None):
        _off = offset if offset is not None else 0
        _len = length if length is not None else len(buff)
        if _off < 0 or _len < 0 or (_off + _len) > len(buff):
            raise IndexError()
        elif length == 0:
            return
        self._ensure_available(_len)
        self._buffer[self._pos : self._pos + _len] = buff
        self._pos += _len

    def write_boolean(self, boolean):
        self._write(1 if boolean else 0)

    def write_byte(self, val):
        self._write(val)

    def write_short(self, val):
        self._ensure_available(SHORT_SIZE_IN_BYTES)
        self._FMT_SHORT.pack_into(self._buffer, self._pos, val)
        self._pos += SHORT_SIZE_IN_BYTES

    def write_char(self, val):
        encoded = val.encode(self._CHAR_ENCODING)
        self.write_from(encoded)

    def write_int(self, val, position=None):
        self._ensure_available(INT_SIZE_IN_BYTES)
        if position is None:
            self._FMT_INT.pack_into(self._buffer, self._pos, val)
            self._pos += INT_SIZE_IN_BYTES
        else:
            self._FMT_INT.pack_into(self._buffer, position, val)

    def write_int_big_endian(self, val):
        self._ensure_available(INT_SIZE_IN_BYTES)
        BE_INT.pack_into(self._buffer, self._pos, val)
        self._pos += INT_SIZE_IN_BYTES

    def write_long(self, val):
        self._ensure_available(LONG_SIZE_IN_BYTES)
        self._FMT_LONG.pack_into(self._buffer, self._pos, val)
        self._pos += LONG_SIZE_IN_BYTES

    def write_float(self, val):
        self._ensure_available(FLOAT_SIZE_IN_BYTES)
        self._FMT_FLOAT.pack_into(self._buffer, self._pos, val)
        self._pos += FLOAT_SIZE_IN_BYTES

    def write_double(self, val):
        self._ensure_available(DOUBLE_SIZE_IN_BYTES)
        self._FMT_DOUBLE.pack_into(self._buffer, self._pos, val)
        self._pos += DOUBLE_SIZE_IN_BYTES

    def write_string(self, val):
        if val is None:
            self.write_int(NULL_ARRAY_LENGTH)
        else:
            encoded_data = val.encode("utf-8")
            self.write_int(len(encoded_data))
            self.write_from(encoded_data)

    def write_bytes(self, val):
        n = len(val)
        self._ensure_available(n)
        for b in val:
            self.write_byte(ord(b))

    def write_chars(self, val):
        n = len(val)
        self._ensure_available(n * CHAR_SIZE_IN_BYTES)
        for c in val:
            self.write_char(c)

    def write_byte_array(self, val):
        _len = len(val) if val is not None else NULL_ARRAY_LENGTH
        self.write_int(_len)
        if _len > 0:
            self.write_from(val)

    def write_boolean_array(self, val):
        self._write_array_fnc(val, self.write_boolean)

    def write_char_array(self, val):
        self._write_array_fnc(val, self.write_char)

    def write_int_array(self, val):
        self._write_array_fnc(val, self.write_int)

    def write_long_array(self, val):
        self._write_array_fnc(val, self.write_long)

    def write_double_array(self, val):
        self._write_array_fnc(val, self.write_double)

    def write_float_array(self, val):
        self._write_array_fnc(val, self.write_float)

    def write_short_array(self, val):
        self._write_array_fnc(val, self.write_short)

    def write_string_array(self, val):
        self._write_array_fnc(val, self.write_string)

    def write_object(self, val):
        self._service.write_object(self, val)

    def to_byte_array(self):
        return self._buffer[: self._pos]

    def get_byte_order(self):
        if self._is_big_endian:
            return "BIG_ENDIAN"
        return "LITTLE_ENDIAN"

    def is_big_endian(self):
        return self._is_big_endian

    def position(self):
        return self._pos

    def set_position(self, position):
        self._pos = position

    def write_zero_bytes(self, count):
        for _ in range(0, count):
            self._write(0)

    def write_utf(self, val):
        self.write_string(val)

    def write_utf_array(self, val):
        self.write_string_array(val)

    # HELPERS
    def _write_array_fnc(self, val, item_write_fnc):
        _len = len(val) if val is not None else NULL_ARRAY_LENGTH
        self.write_int(_len)
        if _len > 0:
            for item in val:
                item_write_fnc(item)

    def _ensure_available(self, length):
        if self._available() < length:
            buffer_length = len(self._buffer)
            new_length = max(buffer_length << 1, buffer_length + length)
            new_buffer = bytearray(new_length)
            new_buffer[: self._pos] = self._buffer[: self._pos]
            self._buffer = new_buffer

    def _available(self):
        return len(self._buffer) - self._pos

    def __repr__(self):
        from binascii import hexlify

        buf = hexlify(self._buffer)
        pos_ = self._pos * 2
        return buf[:pos_] + "[" + buf[pos_] + "]" + buf[pos_ + 1 :]


class EmptyObjectDataOutput(ObjectDataOutput):
    def write_string_array(self, val):
        pass

    def write_string(self, val):
        pass

    def write_short_array(self, val):
        pass

    def write_short(self, val):
        pass

    def write_object(self, val):
        pass

    def write_long_array(self, val):
        pass

    def write_long(self, val):
        pass

    def write_int_array(self, val):
        pass

    def write_int(self, val):
        pass

    def write_from(self, buff, offset=None, length=None):
        pass

    def write_float_array(self, val):
        pass

    def write_float(self, val):
        pass

    def write_double_array(self, val):
        pass

    def write_double(self, val):
        pass

    def write_chars(self, val):
        pass

    def write_char_array(self, val):
        pass

    def write_char(self, val):
        pass

    def write_bytes(self, string):
        pass

    def write_byte_array(self, val):
        pass

    def write_byte(self, val):
        pass

    def write_boolean_array(self, val):
        pass

    def write_boolean(self, val):
        pass

    def to_byte_array(self):
        raise NotImplementedError("to_byte_array not implemented")

    def get_byte_order(self):
        pass

    def write_utf(self, val):
        pass

    def write_utf_array(self, val):
        pass



from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, BEGIN_FRAME_BUF
from hazelcast.protocol.builtin import ListIntegerCodec
from hazelcast.protocol import AnchorDataListHolder
from hazelcast.protocol.builtin import EntryListCodec
from hazelcast.protocol.builtin import DataCodec


class AnchorDataListHolderCodec(object):
    @staticmethod
    def encode(buf, anchor_data_list_holder, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        ListIntegerCodec.encode(buf, anchor_data_list_holder.anchor_page_list)
        EntryListCodec.encode(buf, anchor_data_list_holder.anchor_data_list, DataCodec.encode, DataCodec.encode)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        anchor_page_list = ListIntegerCodec.decode(msg)
        anchor_data_list = EntryListCodec.decode(msg, DataCodec.decode, DataCodec.decode)
        CodecUtil.fast_forward_to_end_frame(msg)
        return AnchorDataListHolder(anchor_page_list, anchor_data_list)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE, EVENT_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x040200
_REQUEST_MESSAGE_TYPE = 262656
# hex: 0x040201
_RESPONSE_MESSAGE_TYPE = 262657
# hex: 0x040202
_EVENT_TOPIC_MESSAGE_TYPE = 262658

_REQUEST_LOCAL_ONLY_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_LOCAL_ONLY_OFFSET + BOOLEAN_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE
_EVENT_TOPIC_PUBLISH_TIME_OFFSET = EVENT_HEADER_SIZE
_EVENT_TOPIC_UUID_OFFSET = _EVENT_TOPIC_PUBLISH_TIME_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(name, local_only):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_LOCAL_ONLY_OFFSET, local_only)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)


def handle(msg, handle_topic_event=None):
    message_type = msg.get_message_type()
    if message_type == _EVENT_TOPIC_MESSAGE_TYPE and handle_topic_event is not None:
        initial_frame = msg.next_frame()
        publish_time = FixSizedTypesCodec.decode_long(initial_frame.buf, _EVENT_TOPIC_PUBLISH_TIME_OFFSET)
        uuid = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _EVENT_TOPIC_UUID_OFFSET)
        item = DataCodec.decode(msg)
        handle_topic_event(item, publish_time, uuid)
        return



from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, EVENT_HEADER_SIZE
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.codec.custom.member_info_codec import MemberInfoCodec
from hazelcast.protocol.builtin import EntryListUUIDListIntegerCodec

# hex: 0x000300
_REQUEST_MESSAGE_TYPE = 768
# hex: 0x000301
_RESPONSE_MESSAGE_TYPE = 769
# hex: 0x000302
_EVENT_MEMBERS_VIEW_MESSAGE_TYPE = 770
# hex: 0x000303
_EVENT_PARTITIONS_VIEW_MESSAGE_TYPE = 771

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE
_EVENT_MEMBERS_VIEW_VERSION_OFFSET = EVENT_HEADER_SIZE
_EVENT_PARTITIONS_VIEW_VERSION_OFFSET = EVENT_HEADER_SIZE


def encode_request():
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    return OutboundMessage(buf, False)


def handle(msg, handle_members_view_event=None, handle_partitions_view_event=None):
    message_type = msg.get_message_type()
    if message_type == _EVENT_MEMBERS_VIEW_MESSAGE_TYPE and handle_members_view_event is not None:
        initial_frame = msg.next_frame()
        version = FixSizedTypesCodec.decode_int(initial_frame.buf, _EVENT_MEMBERS_VIEW_VERSION_OFFSET)
        member_infos = ListMultiFrameCodec.decode(msg, MemberInfoCodec.decode)
        handle_members_view_event(version, member_infos)
        return
    if message_type == _EVENT_PARTITIONS_VIEW_MESSAGE_TYPE and handle_partitions_view_event is not None:
        initial_frame = msg.next_frame()
        version = FixSizedTypesCodec.decode_int(initial_frame.buf, _EVENT_PARTITIONS_VIEW_VERSION_OFFSET)
        partitions = EntryListUUIDListIntegerCodec.decode(msg)
        handle_partitions_view_event(version, partitions)
        return



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec
from hazelcast.protocol.builtin import StringCodec

# hex: 0x1E0200
_REQUEST_MESSAGE_TYPE = 1966592
# hex: 0x1E0201
_RESPONSE_MESSAGE_TYPE = 1966593

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(group_id, service_name, object_name):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    RaftGroupIdCodec.encode(buf, group_id)
    StringCodec.encode(buf, service_name)
    StringCodec.encode(buf, object_name, True)
    return OutboundMessage(buf, True)



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import EntryListCodec

# hex: 0x013100
_REQUEST_MESSAGE_TYPE = 78080
# hex: 0x013101
_RESPONSE_MESSAGE_TYPE = 78081

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(name, entry_processor, predicate):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, entry_processor)
    DataCodec.encode(buf, predicate, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    msg.next_frame()
    return EntryListCodec.decode(msg, DataCodec.decode, DataCodec.decode)



from hazelcast.serialization.api import *
from hazelcast.serialization.bits import *
from hazelcast import six
from hazelcast.six.moves import range


class _ObjectDataInput(ObjectDataInput):
    def __init__(self, buff, offset=0, serialization_service=None, is_big_endian=True):
        self._buffer = buff
        self._service = serialization_service
        self._is_big_endian = is_big_endian
        self._pos = offset
        self._size = len(buff)
        # Local cache struct formats according to endianness
        self._FMT_INT8 = BE_INT8 if self._is_big_endian else LE_INT8
        self._FMT_UINT8 = BE_UINT8 if self._is_big_endian else LE_UINT8
        self._FMT_INT = BE_INT if self._is_big_endian else LE_INT
        self._FMT_SHORT = BE_INT16 if self._is_big_endian else LE_INT16
        self._FMT_CHAR = BE_UINT16 if self._is_big_endian else LE_UINT16
        self._FMT_LONG = BE_LONG if self._is_big_endian else LE_LONG
        self._FMT_FLOAT = BE_FLOAT if self._is_big_endian else LE_FLOAT
        self._FMT_DOUBLE = BE_DOUBLE if self._is_big_endian else LE_DOUBLE

    def read_into(self, buff, offset=None, length=None):
        _off = offset if offset is not None else 0
        _len = length if length is not None else len(buff)
        if _off < 0 or _len < 0 or (_off + _len) > len(self._buffer):
            raise IndexError()
        elif _len == 0:
            return
        if self._pos > self._size:
            raise IndexError()
        if self._pos + _len > self._size:
            _len = self._size - self._pos
        buff[_off : _off + _len] = self._buffer[self._pos : self._pos + _len]
        self._pos += _len

    def skip_bytes(self, count):
        if count <= 0:
            return 0

        if self._pos + count > self._size:
            count = self._size - self._pos

        self._pos += count
        return count

    def read_boolean(self, position=None):
        return self.read_byte(position) != 0

    def read_byte(self, position=None):
        self._check_available(self._pos, BYTE_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_INT8, BYTE_SIZE_IN_BYTES, position)

    def read_unsigned_byte(self, position=None):
        self._check_available(self._pos, BYTE_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_UINT8, BYTE_SIZE_IN_BYTES, position)

    def read_char(self, position=None):
        char_ord = self.read_short(position)
        return six.unichr(char_ord)

    def read_short(self, position=None):
        self._check_available(self._pos, SHORT_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_SHORT, SHORT_SIZE_IN_BYTES, position)

    def read_unsigned_short(self, position=None):
        self._check_available(self._pos, SHORT_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_CHAR, SHORT_SIZE_IN_BYTES, position)

    def read_int(self, position=None):
        self._check_available(position, INT_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_INT, INT_SIZE_IN_BYTES, position)

    def read_long(self, position=None):
        self._check_available(self._pos, LONG_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_LONG, LONG_SIZE_IN_BYTES, position)

    def read_float(self, position=None):
        self._check_available(self._pos, FLOAT_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_FLOAT, FLOAT_SIZE_IN_BYTES, position)

    def read_double(self, position=None):
        self._check_available(self._pos, DOUBLE_SIZE_IN_BYTES)
        return self._read_from_buff(self._FMT_DOUBLE, DOUBLE_SIZE_IN_BYTES, position)

    def read_string(self):
        length = self.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        result = bytearray(length)
        if length > 0:
            self.read_into(result, 0, length)
        return result.decode("utf-8")

    def read_byte_array(self):
        length = self.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        result = bytearray(length)
        if length > 0:
            self.read_into(result, 0, length)

        return result

    def read_boolean_array(self):
        return self._read_array_fnc(self.read_boolean)

    def read_char_array(self):
        return self._read_array_fnc(self.read_char)

    def read_int_array(self):
        return self._read_array_fnc(self.read_int)

    def read_long_array(self):
        return self._read_array_fnc(self.read_long)

    def read_double_array(self):
        return self._read_array_fnc(self.read_double)

    def read_float_array(self):
        return self._read_array_fnc(self.read_float)

    def read_short_array(self):
        return self._read_array_fnc(self.read_short)

    def read_string_array(self):
        return self._read_array_fnc(self.read_string)

    def read_object(self):
        return self._service.read_object(self)

    def is_big_endian(self):
        return self._is_big_endian

    def position(self):
        return self._pos

    def set_position(self, position):
        self._pos = position

    def size(self):
        return self._size

    def get_byte_order(self):
        if self._is_big_endian:
            return "BIG_ENDIAN"
        return "LITTLE_ENDIAN"

    def read_utf(self):
        return self.read_string()

    def read_utf_array(self):
        return self.read_string_array()

    # HELPERS
    def _check_available(self, position, size):
        _position = self._pos if position is None else position
        if _position < 0:
            raise ValueError
        if self._size - _position < size:
            raise EOFError("Cannot read %s bytes!" % size)

    def _read_from_buff(self, fmt, size, position=None):
        if position is None:
            val = fmt.unpack_from(self._buffer, self._pos)
            self._pos += size
        else:
            val = fmt.unpack_from(self._buffer, position)
        return val[0]

    def _read_array_fnc(self, read_item_fnc):
        length = self.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        if length > 0:
            return [read_item_fnc() for _ in range(0, length)]
        return []

    def __repr__(self):
        from binascii import hexlify

        buf = hexlify(self._buffer)
        pos_ = self._pos * 2
        return buf[:pos_] + "[" + buf[pos_] + "]" + buf[pos_ + 1 :]



from hazelcast.serialization.api import IdentifiedDataSerializable
from hazelcast.util import IterationType, get_attr_name

PREDICATE_FACTORY_ID = -20


class Predicate(object):
    """Represents a map entry predicate. Implementations of this class are
    basic building blocks for performing queries on map entries.

    **Special Attributes**

    The predicates that accept an attribute name support two special
    attributes:

    - ``__key`` - instructs the predicate to act on the key associated
      with an item.
    - ``this`` - instructs the predicate to act on the value associated
      with an item.

    **Attribute Paths**

    Dot notation may be used for attribute name to instruct the predicate to
    act on the attribute located at deeper level of an item. Given
    ``"full_name.first_name"`` path the predicate will act on ``first_name``
    attribute of the value fetched by ``full_name`` attribute from the item
    itself. If any of the attributes along the path can't be resolved,
    ``IllegalArgumentError`` will be thrown. Reading of any attribute from
    ``None`` will produce ``None`` value.

    Square brackets notation may be used to instruct the predicate to act on
    the list element at the specified index. Given ``"names[0]"`` path the
    predicate will act on the first item of the list fetched by ``names``
    attribute from the item. The index must be non-negative, otherwise
    ``IllegalArgumentError`` will be thrown. Reading from the index pointing
    beyond the end of the list will produce ``None`` value.

    Special ``any`` keyword may be used to act on every list element. Given
    ``"names[any].full_name.first_name"`` path the predicate will act on
    ``first_name`` attribute of the value fetched by ``full_name`` attribute
    from every list element stored in the item itself under ``names``
    attribute.

    **Handling of None**

    The predicates that accept ``None`` as a value to compare with or a pattern
    to match against if and only if that is explicitly stated in the method
    documentation. In this case, the usual equality logic applies: if ``None``
    is provided, the predicate passes an item if and only if the value stored
    under the item attribute in question is also ``None``.

    Special care must be taken while comparing with ``None`` values *stored*
    inside items being filtered through the predicates created by the following
    methods: :func:`greater`, :func:`greater_or_equal`, :func:`less`,
    :func:`less_or_equal`, :func:`between`. They always evaluate to ``False``
    and therefore never pass such items.

    **Implicit Type Conversion**

    If the type of the stored value doesn't match the type of the value
    provided to the predicate, implicit type conversion is performed before
    predicate evaluation. The provided value is converted to match the type of
    the stored attribute value. If no conversion matching the type exists,
    ``IllegalArgumentError`` is thrown.
    """

    pass


class PagingPredicate(Predicate):
    """This class is a special Predicate which helps to get a page-by-page
    result of a query.

    It can be constructed with a page-size, an inner predicate for filtering,
    and a comparator for sorting. This class is not thread-safe and stateless.
    To be able to reuse for another query, one should call :func:`reset`.
    """

    def reset(self):
        """Resets the predicate for reuse."""
        raise NotImplementedError("reset")

    def next_page(self):
        """Sets page index to next page.

        If new index is out of range, the query results that this
        paging predicate will retrieve will be an empty list.

        Returns:
            int: Updated page index
        """
        raise NotImplementedError("next_page")

    def previous_page(self):
        """Sets page index to previous page.

        If current page index is 0, this method does nothing.

        Returns:
            int: Updated page index.
        """
        raise NotImplementedError("previous_page")

    @property
    def page(self):
        """The current page index.


        :getter: Returns the current page index.
        :setter: Sets the current page index. If the page is out of range, the
            query results that this paging predicate will retrieve will be an
            empty list. New page index must be greater than or equal to ``0``.
        :type: int
        """
        raise NotImplementedError("page")

    @page.setter
    def page(self, new_page):
        raise NotImplementedError("page.setter")

    @property
    def page_size(self):
        """The page size.

        :getter: Returns the page size.
        :type: int
        """
        raise NotImplementedError("page_size")


class _AbstractPredicate(IdentifiedDataSerializable, Predicate):
    def write_data(self, object_data_output):
        raise NotImplementedError("write_data")

    def read_data(self, object_data_input):
        pass

    def get_factory_id(self):
        return PREDICATE_FACTORY_ID

    def get_class_id(self):
        return self.CLASS_ID


class _SqlPredicate(_AbstractPredicate):
    CLASS_ID = 0

    def __init__(self, expression=None):
        self.sql = expression

    def write_data(self, output):
        output.write_string(self.sql)

    def __repr__(self):
        return "SqlPredicate(sql='%s')" % self.sql


class _AndPredicate(_AbstractPredicate):
    CLASS_ID = 1

    def __init__(self, predicates):
        self.predicates = predicates

    def write_data(self, output):
        output.write_int(len(self.predicates))
        for predicate in self.predicates:
            output.write_object(predicate)

    def __repr__(self):
        return "AndPredicate(%s)" % ", ".join([str(p) for p in self.predicates])


class _BetweenPredicate(_AbstractPredicate):
    CLASS_ID = 2

    def __init__(self, attribute, from_, to):
        self.attribute = attribute
        self.from_ = from_
        self.to = to

    def write_data(self, output):
        output.write_string(self.attribute)
        output.write_object(self.to)
        output.write_object(self.from_)

    def __repr__(self):
        return "BetweenPredicate(attribute='%s', from=%s, to=%s)" % (
            self.attribute,
            self.from_,
            self.to,
        )


class _EqualPredicate(_AbstractPredicate):
    CLASS_ID = 3

    def __init__(self, attribute, value):
        self.attribute = attribute
        self.value = value

    def write_data(self, output):
        output.write_string(self.attribute)
        output.write_object(self.value)

    def __repr__(self):
        return "EqualPredicate(attribute='%s', value=%s)" % (self.attribute, self.value)


class _GreaterLessPredicate(_AbstractPredicate):
    CLASS_ID = 4

    def __init__(self, attribute, value, is_equal, is_less):
        self.attribute = attribute
        self.value = value
        self.is_equal = is_equal
        self.is_less = is_less

    def write_data(self, output):
        output.write_string(self.attribute)
        output.write_object(self.value)
        output.write_boolean(self.is_equal)
        output.write_boolean(self.is_less)

    def __repr__(self):
        return "GreaterLessPredicate(attribute='%s', value=%s, is_equal=%s, is_less=%s)" % (
            self.attribute,
            self.value,
            self.is_equal,
            self.is_less,
        )


class _LikePredicate(_AbstractPredicate):
    CLASS_ID = 5

    def __init__(self, attribute, pattern):
        self.attribute = attribute
        self.expression = pattern

    def write_data(self, output):
        output.write_string(self.attribute)
        output.write_string(self.expression)

    def __repr__(self):
        return "LikePredicate(attribute='%s', expression='%s')" % (self.attribute, self.expression)


class _ILikePredicate(_LikePredicate):
    CLASS_ID = 6

    def __repr__(self):
        return "ILikePredicate(attribute='%s', expression='%s')" % (self.attribute, self.expression)


class _InPredicate(_AbstractPredicate):
    CLASS_ID = 7

    def __init__(self, attribute, values):
        self.attribute = attribute
        self.values = values

    def write_data(self, output):
        output.write_string(self.attribute)
        output.write_int(len(self.values))
        for value in self.values:
            output.write_object(value)

    def __repr__(self):
        return "InPredicate(attribute='%s', %s)" % (
            self.attribute,
            ",".join([str(x) for x in self.values]),
        )


class _InstanceOfPredicate(_AbstractPredicate):
    CLASS_ID = 8

    def __init__(self, class_name):
        self.class_name = class_name

    def write_data(self, output):
        output.write_string(self.class_name)

    def __repr__(self):
        return "InstanceOfPredicate(class_name='%s')" % self.class_name


class _NotEqualPredicate(_EqualPredicate):
    CLASS_ID = 9

    def __repr__(self):
        return "NotEqualPredicate(attribute='%s', value=%s)" % (self.attribute, self.value)


class _NotPredicate(_AbstractPredicate):
    CLASS_ID = 10

    def __init__(self, predicate):
        self.predicate = predicate

    def write_data(self, output):
        output.write_object(self.predicate)

    def __repr__(self):
        return "NotPredicate(predicate=%s)" % self.predicate


class _OrPredicate(_AndPredicate):
    CLASS_ID = 11

    def __repr__(self):
        return "OrPredicate(%s)" % ", ".join([str(p) for p in self.predicates])


class _RegexPredicate(_AbstractPredicate):
    CLASS_ID = 12

    def __init__(self, attribute, pattern):
        self.attribute = attribute
        self.pattern = pattern

    def write_data(self, output):
        output.write_string(self.attribute)
        output.write_string(self.pattern)

    def __repr__(self):
        return "RegexPredicate(attribute='%s', pattern='%s')" % (self.attribute, self.pattern)


class _FalsePredicate(_AbstractPredicate):
    CLASS_ID = 13

    def write_data(self, object_data_output):
        pass

    def __repr__(self):
        return "FalsePredicate()"


class _TruePredicate(_AbstractPredicate):
    CLASS_ID = 14

    def write_data(self, object_data_output):
        pass

    def __repr__(self):
        return "TruePredicate()"


class _PagingPredicate(_AbstractPredicate, PagingPredicate):
    CLASS_ID = 15

    def __init__(self, predicate, page_size, comparator=None):
        if isinstance(predicate, PagingPredicate):
            raise TypeError("Nested paging predicate not supported.")

        if page_size <= 0:
            raise ValueError("page_size should be greater than 0.")

        self._internal_predicate = predicate
        self._page_size = page_size
        self._page = 0  # initialized to be on first page
        self.comparator = comparator
        self.iteration_type = IterationType.ENTRY
        self.anchor_list = []  # List of pairs: (nearest page, (anchor key, anchor value))

    def __repr__(self):
        return "PagingPredicate(predicate=%s, page_size=%s, comparator=%s)" % (
            self._internal_predicate,
            self.page_size,
            self.comparator,
        )

    def write_data(self, output):
        output.write_object(self._internal_predicate)
        output.write_object(self.comparator)
        output.write_int(self.page)
        output.write_int(self._page_size)
        output.write_string(get_attr_name(IterationType, self.iteration_type))
        output.write_int(len(self.anchor_list))
        for nearest_page, (anchor_key, anchor_value) in self.anchor_list:
            output.write_int(nearest_page)
            output.write_object(anchor_key)
            output.write_object(anchor_value)

    def next_page(self):
        self.page += 1
        return self.page

    def previous_page(self):
        if self.page != 0:
            self.page -= 1
        return self.page

    def reset(self):
        self.iteration_type = IterationType.ENTRY
        del self.anchor_list[:]
        self.page = 0

    @property
    def page(self):
        return self._page

    @page.setter
    def page(self, new_page):
        if new_page < 0:
            raise ValueError("new_page should be positive or 0.")

        self._page = new_page

    @property
    def page_size(self):
        return self._page_size


def sql(expression):
    """Creates a predicate that will pass items that match the given SQL
    ``where`` expression.

    The following operators are supported: ``=``, ``<``, ``>``, ``<=``, ``>=``
    ``==``, ``!=``, ``<>``, ``BETWEEN``, ``IN``, ``LIKE``, ``ILIKE``, ``REGEX``
    ``AND``, ``OR``, ``NOT``.

    The operators are case-insensitive, but attribute names are case sensitive.

    Example: ``active AND (age > 20 OR salary < 60000)``

    Differences to standard SQL:

    - We don't use ternary boolean logic. ``field=10`` evaluates to ``false``,
      if ``field`` is ``null``, in standard SQL it evaluates to ``UNKNOWN``.
    - ``IS [NOT] NULL`` is not supported, use ``=NULL`` or ``<>NULL``.
    - ``IS [NOT] DISTINCT FROM`` is not supported, but ``=`` and ``<>`` behave
      like it.

    Args:
        expression (str): The ``where`` expression.

    Returns:
        Predicate: The created **sql** predicate instance.
    """
    return _SqlPredicate(expression)


def equal(attribute, value):
    """Creates a predicate that will pass items if the given ``value`` and the
    value stored under the given item ``attribute`` are equal.

    Args:
        attribute (str): The attribute to fetch the value for comparison from.
        value: The value to compare the attribute value against. Can be
            ``None``.

    Returns:
        Predicate: The created **equal** predicate instance.
    """
    return _EqualPredicate(attribute, value)


def not_equal(attribute, value):
    """Creates a predicate that will pass items if the given ``value`` and the
    value stored under the given item ``attribute`` are not equal.

    Args:
        attribute (str): The attribute to fetch the value for comparison from.
        value: The value to compare the attribute value against. Can be
            ``None``.

    Returns:
        Predicate: The created **not equal** predicate instance.
    """
    return _NotEqualPredicate(attribute, value)


def like(attribute, pattern):
    """Creates a predicate that will pass items if the given ``pattern``
    matches the value stored under the given item ``attribute``.

    Args:
        attribute (str): The attribute to fetch the value for matching from.
        pattern (str): The pattern to match the attribute value against.
            The ``%`` (percentage sign) is a placeholder for multiple
            characters, the ``_`` (underscore) is a placeholder for a single
            character. If you need to match the percentage sign or the
            underscore character itself, escape it with the backslash, for
            example ``"\\%"`` string will match the percentage sign. Can be
            ``None``.

    Returns:
        Predicate: The created **like** predicate instance.

    See Also:
        :func:`ilike` and :func:`regex`
    """
    return _LikePredicate(attribute, pattern)


def ilike(attribute, pattern):
    """Creates a predicate that will pass items if the given ``pattern``
    matches  the value stored under the given item ``attribute`` in a
    case-insensitive manner.

    Args:
        attribute (str): The attribute to fetch the value for matching from.
        pattern (str): The pattern to match the attribute value against.
            The ``%`` (percentage sign) is a placeholder for multiple
            characters, the ``_`` (underscore) is a placeholder for a single
            character. If you need to match the percentage sign or the
            underscore character itself, escape it with the backslash, for
            example ``"\\%"`` string will match the percentage sign. Can be
            ``None``.

    Returns:
        Predicate: The created **case-insensitive like** predicate instance.

    See Also:
        :func:`like` and :func:`regex`
    """
    return _ILikePredicate(attribute, pattern)


def regex(attribute, pattern):
    """Creates a predicate that will pass items if the given ``pattern``
    matches the value stored under the given item ``attribute``.

    Args:
        attribute (str): The attribute to fetch the value for matching from.
        pattern (str): The pattern to match the attribute value against. The
            pattern interpreted exactly the same as described in
            https://docs.oracle.com/javase/8/docs/api/java/util/regex/Pattern.html.
            Can be ``None``.

    Returns:
        Predicate: The created **regex** predicate instance.

    See Also:
        :func:`ilike` and :func:`like`
    """
    return _RegexPredicate(attribute, pattern)


def and_(*predicates):
    """Creates a predicate that will perform the logical ``and`` operation on
    the given predicates.

    If no predicate is provided as argument, the created predicate will always
    evaluate to ``true`` and will pass any item.

    Args:
        *predicates (Predicate): The child predicates to form the resulting
            ``and`` predicate from.

    Returns:
        Predicate: The created **and** predicate instance.
    """
    return _AndPredicate(predicates)


def or_(*predicates):
    """Creates a predicate that will perform the logical ``or`` operation on
    the given predicates.

    If no predicate is provided as argument, the created predicate will always
    evaluate to ``false`` and will never pass any items.

    Args:
        *predicates (Predicate): The child predicates to form the resulting
            ``or`` predicate from.

    Returns:
        Predicate: The created **or** predicate instance.
    """
    return _OrPredicate(predicates)


def not_(predicate):
    """Creates a predicate that will negate the result of the given
    ``predicate``.

    Args:
        predicate (Predicate): The predicate to negate the value of.

    Returns:
        Predicate: The created **not** predicate instance.
    """
    return _NotPredicate(predicate)


def between(attribute, from_, to):
    """Creates a predicate that will pass items if the value stored under the
    given item ``attribute`` is contained inside the given range.

    The range begins at the given ``from_`` bound and ends at the given ``to``
    bound. The bounds are inclusive.

    Args:
        attribute (str): The attribute to fetch the value to check from.
        from_: The inclusive lower bound of the range to check.
        to: The inclusive upper bound of the range to check.

    Returns:
        Predicate: The created **between** predicate.
    """
    return _BetweenPredicate(attribute, from_, to)


def in_(attribute, *values):
    """Creates a predicate that will pass items if the value stored under the
    given item ``attribute`` is a member of the given ``values``.

    Args:
        attribute (str): The attribute to fetch the value to test from.
        *values: The values set to test the membership in. Individual values
            can be ``None``.

    Returns:
        Predicate: The created **in** predicate.
    """
    return _InPredicate(attribute, values)


def instance_of(class_name):
    """Creates a predicate that will pass entries for which the value class is
    an instance of the given ``class_name``.

    Args:
        class_name (str): The name of class the created predicate will check
            for.

    Returns:
        Predicate: The created **instance of** predicate.
    """
    return _InstanceOfPredicate(class_name)


def false():
    """Creates a predicate that will filter out all items.

    Returns:
        Predicate: The created **false** predicate.
    """
    return _FalsePredicate()


def true():
    """Creates a predicate that will pass all items.

    Returns:
        Predicate: The created **true** predicate.
    """
    return _TruePredicate()


def paging(predicate, page_size, comparator=None):
    """Creates a paging predicate with an inner predicate, page size and
    comparator. Results will be filtered via inner predicate and will be
    ordered via comparator if provided.

    Args:
        predicate (Predicate): The inner predicate through which results will
            be filtered. Can be ``None``. In that case, results will not be
            filtered.
        page_size (int): The page size.
        comparator (hazelcast.serialization.api.Portable or hazelcast.serialization.api.IdentifiedDataSerializable):
            The comparator through which results will be ordered. The
            comparision logic must be defined on the server side. Can be
            ``None``. In that case, the results will be returned in natural
            order.

    Returns:
        PagingPredicate: The created **paging** predicate.
    """
    return _PagingPredicate(predicate, page_size, comparator)


def greater(attribute, value):
    """
    Creates a predicate that will pass items if the value stored under the
    given item ``attribute`` is greater than the given ``value``.

    Args:
        attribute (str): The left-hand side attribute to fetch the value for
            comparison from.
        value: The right-hand side value to compare the attribute value
            against.

    Returns:
        Predicate: The created **greater than** predicate.
    """
    return _GreaterLessPredicate(attribute, value, False, False)


def greater_or_equal(attribute, value):
    """Creates a predicate that will pass items if the value stored under the
    given item ``attribute`` is greater than or equal to the given ``value``.

    Args:
        attribute (str): the left-hand side attribute to fetch the value for
            comparison from.
        value: The right-hand side value to compare the attribute value
            against.

    Returns:
        Predicate: The created **greater than or equal to** predicate.
    """
    return _GreaterLessPredicate(attribute, value, True, False)


def less(attribute, value):
    """Creates a predicate that will pass items if the value stored under the
    given item ``attribute`` is less than the given ``value``.

    Args:
        attribute (str): The left-hand side attribute to fetch the value for
            comparison from.
        value: The right-hand side value to compare the attribute value
            against.

    Returns:
        Predicate: The created **less than** predicate.
    """
    return _GreaterLessPredicate(attribute, value, False, True)


def less_or_equal(attribute, value):
    """Creates a predicate that will pass items if the value stored under the
    given item ``attribute`` is less than or equal to the given ``value``.

    Args:
        attribute (str): The left-hand side attribute to fetch the value for
            comparison from.
        value: The right-hand side value to compare the attribute value
            against.

    Returns:
        Predicate: The created **less than or equal to** predicate.
    """
    return _GreaterLessPredicate(attribute, value, True, True)



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.protocol import PagingPredicateHolder
from hazelcast.protocol.codec.custom.anchor_data_list_holder_codec import AnchorDataListHolderCodec
from hazelcast.protocol.builtin import DataCodec

_PAGE_SIZE_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_PAGE_SIZE_DECODE_OFFSET = 0
_PAGE_ENCODE_OFFSET = _PAGE_SIZE_ENCODE_OFFSET + INT_SIZE_IN_BYTES
_PAGE_DECODE_OFFSET = _PAGE_SIZE_DECODE_OFFSET + INT_SIZE_IN_BYTES
_ITERATION_TYPE_ID_ENCODE_OFFSET = _PAGE_ENCODE_OFFSET + INT_SIZE_IN_BYTES
_ITERATION_TYPE_ID_DECODE_OFFSET = _PAGE_DECODE_OFFSET + INT_SIZE_IN_BYTES
_INITIAL_FRAME_SIZE = _ITERATION_TYPE_ID_ENCODE_OFFSET + BYTE_SIZE_IN_BYTES - 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS


class PagingPredicateHolderCodec(object):
    @staticmethod
    def encode(buf, paging_predicate_holder, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_int(initial_frame_buf, _PAGE_SIZE_ENCODE_OFFSET, paging_predicate_holder.page_size)
        FixSizedTypesCodec.encode_int(initial_frame_buf, _PAGE_ENCODE_OFFSET, paging_predicate_holder.page)
        FixSizedTypesCodec.encode_byte(initial_frame_buf, _ITERATION_TYPE_ID_ENCODE_OFFSET, paging_predicate_holder.iteration_type_id)
        buf.extend(initial_frame_buf)
        AnchorDataListHolderCodec.encode(buf, paging_predicate_holder.anchor_data_list_holder)
        CodecUtil.encode_nullable(buf, paging_predicate_holder.predicate_data, DataCodec.encode)
        CodecUtil.encode_nullable(buf, paging_predicate_holder.comparator_data, DataCodec.encode)
        CodecUtil.encode_nullable(buf, paging_predicate_holder.partition_key_data, DataCodec.encode)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        page_size = FixSizedTypesCodec.decode_int(initial_frame.buf, _PAGE_SIZE_DECODE_OFFSET)
        page = FixSizedTypesCodec.decode_int(initial_frame.buf, _PAGE_DECODE_OFFSET)
        iteration_type_id = FixSizedTypesCodec.decode_byte(initial_frame.buf, _ITERATION_TYPE_ID_DECODE_OFFSET)
        anchor_data_list_holder = AnchorDataListHolderCodec.decode(msg)
        predicate_data = CodecUtil.decode_nullable(msg, DataCodec.decode)
        comparator_data = CodecUtil.decode_nullable(msg, DataCodec.decode)
        partition_key_data = CodecUtil.decode_nullable(msg, DataCodec.decode)
        CodecUtil.fast_forward_to_end_frame(msg)
        return PagingPredicateHolder(anchor_data_list_holder, predicate_data, comparator_data, page_size, page, iteration_type_id, partition_key_data)
